
//*****************************************************************************}
//                                                                             }
//       Password Manager plugin                                               }
//       Autofill Engine for Gecko-Engine WebBrowsers                          }
//                                                                             }
//       (C) 1997-2012 Kaspersky Lab ZAO                                       }
//                                                                             }
//*****************************************************************************}

// ==UserScript==
// @include *
// @include http://*
// @include https://*
// ==/UserScript==

// >> Content Script ===========================================================

// debug modes
const kpmDebug_Errors = true;
const kpmDebug_LogInfo = false;

// kpmLog -----------------------------------------------------------------------

const kpmLog =
{
  logMessage: function(text)
  {
    try 
    {
      opera.postError('Kaspersky PURE: ' + text); // TODO: Need replace to "logMessage" function for Opera!
    }
    catch (ErrorMessage)
    {
      // keep silence
    }
  },
  
  logError: function(text)
  {
    if (kpmDebug_Errors)
    try
    {
      opera.postError('Kaspersky PURE: ' + text);
    }
    catch (ErrorMessage)
    {
      // keep silence
    }
  }
}

// kpmAutofill ------------------------------------------------------------------

var kpmAutofill__ = null;
var kpmAutofillCheckCount = 0;
const kpmAutofillMaxCheckCount = 15;

function kpmGetAutofill()
{
  try
  {
    // 0. init ADocument & ABody variables
    var ADocument = null;
    var ABody = null;
    var BodyElements = document.getElementsByTagName('body');
    if (BodyElements && BodyElements.length)
    {
      // document has <body> tag
      ADocument = document;
      ABody = BodyElements[0];
    }
    else
    {
      // it seems the document has <frameset> tag
      var frames = document.getElementsByTagName('frame');
      for(var i = 0; i < frames.length; i++)
      {
        try
        {
          var w = frames[i];
          var d = w.contentDocument;
          BodyElements = d.getElementsByTagName('body');
          if (BodyElements && BodyElements.length)
          {
            ADocument = d;
            ABody = BodyElements[0];
            break;
          }
        }
        catch (ErrorMessage)
        {
          // error accessing the frame!
        }
      }
    }
    if (!ADocument || !ABody)
      return;
    
    // 1. add plugin
    if (kpmAutofill__ == null || ADocument.getElementById('kpmAutofillPlugin') == null)
    {
      kpmAutofill__ = ADocument.createElement('embed');
      kpmAutofill__.setAttribute('id', 'kpmAutofillPlugin');
      kpmAutofill__.setAttribute('type', 'application/x-npkpmautofillplugin');
      kpmAutofill__.setAttribute('style', 'visibility: hidden; width: 0; height: 0; position: absolute;');
      ABody.appendChild(kpmAutofill__);
    }
    
    // 2. try to get already added plugin
    kpmAutofill__ = ADocument.getElementById('kpmAutofillPlugin');
  }
  catch (ErrorMessage)
  {
    kpmAutofill__ = null;
    kpmLog.logError('kpmGetAutofill() Error: ' + ErrorMessage);
  }
}

function kpmAutofill()
{
  // 1. check if kpmAutofill NPAPI initialized?
  if (kpmAutofill__)
  {
    if (kpmAutofillCheckCount >= kpmAutofillMaxCheckCount)
    {
      // return empty autofill plugin cause attempts count are exceeded
      return null;
    }
    try
    {
      kpmAutofill__.IsPasswordManagerUnlocked();
    }
    catch (ErrorMessage)
    {
      // any error occured accessing our plugin (web-page can be reloaded, etc.)
      // we need to reinitialize the plugin
      kpmAutofill__ = null;
      kpmAutofillCheckCount++;
    }
  }
  // 2. try to initialize kpmAutofill NPAPI (if need)
  if (!kpmAutofill__)
  {
    kpmGetAutofill();
  }
  // 3. return the kpmAutofill NPAPI
  return kpmAutofill__;
}

// Events engine ---------------------------------------------------------------

// add base events
window.addEventListener('load', kpmOnLoad, true);
window.addEventListener('submit', kpmOnSubmit_Common, true);
window.addEventListener('kpm_submit', kpmOnSubmit_JS, false, true);
window.addEventListener('beforeunload', kpmOnBeforeUnload, true);
addEventListener('unload', kpmOnUnload, false);

function kpmRemoveBaseEvents()
{
  window.removeEventListener('load', kpmOnLoad, true);
  window.removeEventListener('submit', kpmOnSubmit_Common, true);
  window.removeEventListener('kpm_submit', kpmOnSubmit_JS, false);
  window.removeEventListener('beforeunload', kpmOnBeforeUnload, true);
  removeEventListener('unload', kpmOnUnload, false);
}

function kpmAddAdditionalEvents()
{
  // empty
}

function kpmRemoveAdditionalEvents()
{
  // empty
}

// Rescue/Remove Passwords engine ----------------------------------------------

function kpmProcessOnLoad(ADocument, Url)
{
  return false;
}

// Content script / Background page exchange engine ----------------------------

opera.extension.addEventListener('message', kpmOnMessage, false);
// query AuthKey from background script
opera.extension.postMessage({name: 'kpmGetAuthKey'});

function kpmOnMessage(AMessage)
{
  // 1. process message
  try
  {
    if (AMessage)
      kpmProcessContentMessage(AMessage.data);
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('Error in onMessage event: ' + ErrorMessage);
  }
}

// << Content Script ===========================================================

// >> Strings Script ===========================================================

// log modes
var kpmLog_WideIsEmpty = false;
var kpmLog_WideSameText = false;
var kpmLog_WideSearchText = false;

const kpmLineBreak = '\r\n';
 
function kpmWidePrepareStr(Text)
{
  var Result;
  try
  {
    if (Text && Text.value != undefined)
      Result = new String(Text.value); // String object
    else
      Result = new String(Text);       // String
  }
  catch (ErrorMessage)
  {
    if (kpmDebug_Errors)
      kpmLog.logError('kpmWidePrepareStr() Error: ' + ErrorMessage);
    Result = new String(Text); // String
  }
  return Result;
}

function kpmWidePrepareLowerCaseStr(Text)
{
  var Result = kpmWidePrepareStr(Text);
  Result = Result.toLowerCase();
  return Result;  
}

function kpmWidePrepareUpperCaseStr(Text)
{
  var Result = kpmWidePrepareStr(Text);
  Result = Result.toUpperCase();
  return Result;  
}

function kpmWideIsEmpty(Text)
{
  try
  {
    if (typeof Text == 'object')
    {
      if (Text == null)
      {
        if (kpmLog_WideIsEmpty)
        {
          kpmLog.logMessage('kpmWideIsEmpty() Text[object]: <' + Text + '/' + Text.value + '>');
          kpmLog.logMessage('kpmWideIsEmpty() Text == null');
        }
        return true;
      }
      if ((Text.value == '') || (Text.value == undefined) || (Text.empty && Text.empty()))
      {
        if (kpmLog_WideIsEmpty)
        {
          kpmLog.logMessage('kpmWideIsEmpty() Text[object]: <' + Text + '/' + Text.value + '>');
          kpmLog.logMessage('kpmWideIsEmpty() Text.empty()');
        }
        return true;
      }        
    } else
    if (typeof Text == 'string')
    {      
      if (Text == '' || Text == null || Text == undefined)
      {
        if (kpmLog_WideIsEmpty)
        {
          kpmLog.logMessage('kpmWideIsEmpty() Text[string]: <' + Text + '>');
          kpmLog.logMessage('kpmWideIsEmpty() Text == "" || Text == null || Text == undefined');
        }
        return true;
      }
    } else
    { 
      // unknown type
      if (kpmLog_WideIsEmpty)
        kpmLog.logMessage('kpmWideIsEmpty() Text[' + typeof Text + ']: <' + Text + '>');
      return true;
    }
    return false;
  }
  catch (ErrorMessage)
  {
    if (kpmDebug_Errors)
      kpmLog.logError('kpmWideIsEmpty() Error: ' + ErrorMessage);
  }
  return false;  
}

function kpmWideSameText(Text1, Text2)
{
  try
  {
    var Str1 = kpmWidePrepareLowerCaseStr(Text1);
    var Str2 = kpmWidePrepareLowerCaseStr(Text2);
    if (kpmLog_WideSameText)
      kpmLog.logMessage(
        'Text1: <' + Text1 + '>, Text2: <' + Text2 + '>, ' +
        'Str1: <' + Str1 + '>, Str2: <' + Str2 + '>, Str1 == Str2: ' + (Str1 == Str2)
      );
    return Str1 == Str2;
  }
  catch (ErrorMessage)
  {
    if (kpmDebug_Errors)
      kpmLog.logError('kpmWideSameText() Error: ' + ErrorMessage);
    return false;
  }
}
  
function kpmWideSearchText(ASearchingText, AText)
{
  try
  {
    var SearchingText = kpmWidePrepareLowerCaseStr(ASearchingText);
    var Text = kpmWidePrepareLowerCaseStr(AText);
    if (kpmLog_WideSearchText)
      kpmLog.logMessage(
        'ASearchingText: <' + ASearchingText + '>, AText: <' + AText + '>, ' +
        'SearchingText: <' + SearchingText + '>, Text: <' + Text + '>, Text.indexOf(SearchingText): ' + Text.indexOf(SearchingText));
    if (Text.indexOf(SearchingText) >= 0)
      return true;
    else
      return false;
  }
  catch (ErrorMessage)
  {
    if (kpmDebug_Errors)
      kpmLog.logError('kpmWideSearchText() Error: ' + ErrorMessage);
    return false;
  }
}

function kpmTrim(sString)
{
  if (typeof sString != 'string')
    return sString;
  while (sString.substring(0, 1) == ' ')
    sString = sString.substring(1, sString.length);
  while (sString.substring(sString.length-1, sString.length) == ' ')
    sString = sString.substring(0, sString.length-1);
  return sString;
}

function kpmCompareUrl(AUrl1, AUrl2)
{
  try
  {
    AUrl1 = kpmWidePrepareStr(AUrl1);
    AUrl2 = kpmWidePrepareStr(AUrl2);
    // canonicalize Url
    AUrl1 = AUrl1.replace(/%7C/gi, "|");
    AUrl2 = AUrl2.replace(/%7C/gi, "|");
    // compare Url
    if (kpmWideSameText(AUrl1, AUrl2) ||
        kpmWideSearchText(AUrl1, AUrl2) ||
        kpmWideSearchText(AUrl2, AUrl1)
       )
      return true;
    else
      return false;
  }
  catch (ErrorMessage)
  {
    if (kpmDebug_Errors)
      kpmLog.logError('kpmCompareUrl() Error: ' + ErrorMessage);
    return false;
  }
}

// << Strings Script ===========================================================

// >> Content Core Script ======================================================

var kpmAdditionalEventsAdded = false;
var kpmIsAutofillingDocument = false;
var kpmRescuingPasswords = false;
var kpmRemovingPasswords = false;

// debug modes
const kpmDebug_NotEmptyPasswordControlExistsInForm = false;
const kpmDebug_PasswordControlExists = false;
const kpmDebug_OnLoad = false;
const kpmDebug_Autofill = false;
const kpmDebug_Elements = false;
const kpmDebug_OnSubmit = false;
const kpmDebug_OnTabSelect = false;
const kpmDebug_WebProgressListenerOnStateChange = false;
const kpmDebug_GetActiveElementScreenRect = false;
const kpmDebug_SetActiveElementSelectedText = false;
const kpmDebug_IsMultipageAuthorizationLogin = false;
const kpmDebug_IsAllowedAdditionalFieldElement = false;
const kpmDebug_IsHTMLElementVisible = false;
const kpmDebug_IsHTMLElementVisible_Details = false;
const kpmDebug_ExportDOMToXml = false;
const kpmDebug_FixFocusingEdit = false;
const kpmDebug_InstallHTMLFormElementPrototypeForDocument = false;
 
// AdditionalFields constants
const kpm_szAdditionalFieldsManager = 'AdditionalFieldsManager';
const kpm_szFormSubmitField = 'FormSubmitField';
const kpm_szButtonSubmitField = 'ButtonSubmitField';
const kpm_szButtonCheckAvailabilityField = 'ButtonCheckAvailabilityField';
const kpm_szFocusingField = 'FocusingField';
const kpm_szAdditionalField = 'AdditionalField';
const kpm_szFieldType = 'FieldType';
const kpm_szFieldName = 'FieldName';
const kpm_szFieldValue = 'FieldValue';
const kpm_szAutofillValueType = 'AutofillValueType';
const kpm_szHtmlDocumentUrl = 'HtmlDocumentUrl';
const kpm_szHtmlFormName = 'HtmlFormName';
const kpm_szHtmlFormAction = 'HtmlFormAction';

// FieldType constants
const kpm_szFORM = 'FORM';
const kpm_szA = 'A';
const kpm_szIMG = 'IMG';
const kpm_szINPUT = 'INPUT';
const kpm_szBUTTON = 'BUTTON';
const kpm_szIMAGE = 'IMAGE';
const kpm_szSUBMIT = 'SUBMIT';
const kpm_szPASSWORD = 'PASSWORD';
const kpm_szCHECKBOX = 'CHECKBOX';
const kpm_szRADIO = 'RADIO';
const kpm_szTEXT = 'TEXT';
const kpm_szTEXTAREA = 'TEXTAREA';
const kpm_szSELECT = 'SELECT';
const kpm_szSELECT_MULTIPLE = 'SELECT-MULTIPLE';
const kpm_szSELECT_ONE = 'SELECT-ONE';

const kpm_FieldValueIsTextPrefix = '"';

// TkpmAdditionalFieldType
const kpm_afUnknown = 0;
const kpm_afHtmlForm = 1;
const kpm_afButton = 2;
const kpm_afButtonImage = 3;
const kpm_afButtonSubmit = 4;
const kpm_afEdit = 5;
const kpm_afEditPassword = 6;
const kpm_afCheckBox = 7;
const kpm_afRadioButton = 8;
const kpm_afMemo = 9;
const kpm_afComboBox = 10;
const kpm_afComboBoxDropDownList = 11;
const kpm_afListBox = 12;
const kpm_afAll = 13;

// TkpmAutofillValueType
const kpm_atValue = 0;
const kpm_atLogin = 1;
const kpm_atPassword = 2;

// TkpmAutofillReason
const kpm_arUnknown              = 0;  // nothing to do
const kpm_arAutofill             = 1;  // autofill fields
const kpm_arHighlight            = 2;  // just highlight fields without autofill
const kpm_arAutofillAndHighlight = 3;  // autofill & highlight fields

// THTMLFormMethod
const kpm_htmlFormMethod_Unknown = 0;
const kpm_htmlFormMethod_Get = 1;
const kpm_htmlFormMethod_Post = 2;

function kpmIsRectEmpty(ARect)
{
  return !ARect || (ARect.left == 0 && ARect.right == 0 || ARect.top == 0 && ARect.bottom == 0);
}

function kpmIsEqualRect(ARect1, ARect2)
{
  if (ARect1 && ARect2)
    return ARect1.left   == ARect2.left &&
           ARect1.top    == ARect2.top &&
           ARect1.right  == ARect2.right &&
           ARect1.bottom == ARect2.bottom;
  else
    return false;
}

function kpmCompareIDorName(InputTextElement, AValue)
{
  if (InputTextElement)
  {
    if (kpmWideSearchText(AValue, InputTextElement.id) ||
        kpmWideSearchText(AValue, kpmGetElementName(InputTextElement)))
      return true;
    else
      return false;
  } else
  {
    return false;
  }
}

var
  spAuthKey = '';
  kpmLastFocusedElement = null;
  kpmLastFocusedEditPasswordRect = null;

function kpmCheckBalloons(AWindow, ADocument)
{
  // this function used to show/hide Virtual Keyboard, Password Generator Balloon
  if (kpmAutofill())
  try
  {
    var AllowProcessVirtualKeyboard = 
      kpmAutofill().AllowVirtualKeyboardAutoShow();    
    var AllowProcessPasswordGeneratorBalloon = 
      // deny to process if autofilling the document!
      (!kpmIsAutofillingDocument) &&
      // allow to process if PM unlocked
      kpmAutofill().IsPasswordManagerUnlocked();

    if (AllowProcessVirtualKeyboard ||
        AllowProcessPasswordGeneratorBalloon
       )
    {
      var ActiveElement = kpmGetDocumentFocusedElement(ADocument);      
      var FieldType = kpmGetAdditionalFieldType(ActiveElement);
      
      // is virtual keyboard allowed to process?
      if (AllowProcessVirtualKeyboard)
      {
        // obtain MustAutoShow
        var MustAutoShow = 
          (FieldType == kpm_afEdit) ||
          (FieldType == kpm_afEditPassword) ||
          (FieldType == kpm_afMemo);
        // obtain element rect
        var rcActiveElement = kpmGetElementScreenPosition(ActiveElement);
        if (MustAutoShow)
        {
          // show virtual keyboard
          kpmAutofill().ShowVirtualKeyboard(
            rcActiveElement.left, rcActiveElement.top, 
            rcActiveElement.right, rcActiveElement.bottom);
        }
        else
        {
          // auto-hide virtual keyboard for unsupported element
          kpmAutofill().AutoHideVirtualKeyboard(
            rcActiveElement.left, rcActiveElement.top, 
            rcActiveElement.right, rcActiveElement.bottom);
        }
      } 
      // is password generator balloon allowed to process?
      else if (AllowProcessPasswordGeneratorBalloon)
      {
        // deny to process <body> element 
        //   it's a bug in Opera: when element looses focus - active element changed to body one
        if (ActiveElement)
        {
          if (kpmWideSameText(ActiveElement.tagName, 'BODY'))
            return;
        }
        
        // obtain MustAutoShow, AllowUseHistory
        var MustAutoShow = false;
        var AllowUseHistory = false;
        var rcActiveElement = null;
        if (FieldType == kpm_afEditPassword)
        {
          // obtain element rect
          rcActiveElement = kpmGetElementScreenPosition(ActiveElement);
          // re-show balloon if last focused element's rect changed!
          if (kpmLastFocusedElement == ActiveElement &&
              kpmIsEqualRect(kpmLastFocusedEditPasswordRect, rcActiveElement)
             )
          {
            return;
          }
          kpmLastFocusedEditPasswordRect = rcActiveElement;
          var URL = kpmGetDocumentUrl(ADocument);
          var DOMXml = kpmExportDOMToXml(AWindow);
          var objAllowUseHistory = {};
          MustAutoShow = kpmAutofill().AllowShowPasswordGeneratorBalloon(URL, DOMXml, objAllowUseHistory);
          AllowUseHistory = objAllowUseHistory.value;
        }
        if (MustAutoShow)
        {
          // show password generator balloon
          kpmAutofill().ShowPasswordGeneratorBalloon(
            rcActiveElement.left, rcActiveElement.top, 
            rcActiveElement.right, rcActiveElement.bottom, AllowUseHistory);
        }
        else if (kpmLastFocusedElement != ActiveElement)
        {
          // auto-hide password generator balloon
          kpmAutofill().AutoHidePasswordGeneratorBalloon();
        }
        kpmLastFocusedElement = ActiveElement;
      }
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmCheckVirtualKeyboard() Error: ' + ErrorMessage);
  }
}

function kpmAutofillEngineOnTimerFocusedWindow()
{
  if (kpmAutofill())
  try
  {
    var AllowProceed = true;
    var objMessageID = {};
    // check Window Content
    var AWindow = kpmWindowContent(window);
    if (!AWindow)
      return;
    var ADocument = AWindow.document;
    
    // show/hide Virtual Keyboard, Password Generator Balloon
    kpmCheckBalloons(AWindow, ADocument);
    
    var AllowProceedDocument = kpmAllowProceedDocument(ADocument, false);
    // check if we need to export DOM to XML
    if (AllowProceed &&
        kpmAutofill().MustGetDOMXml(objMessageID)
       )
    try
    {
      if (!AllowProceedDocument)
      {
        kpmAutofill().MessageProcessFailed(objMessageID.value);
        if (kpmDebug_LogInfo)
          kpmLog.logMessage('MustGetDOMXml() Document process denied for <' + kpmGetDocumentUrl(ADocument) + '>');
      }
      else
      {
        AllowProceed = false;
        if (kpmDebug_LogInfo)
          kpmLog.logMessage('MustGetDOMXml()');
        var DOMXml = kpmExportDOMToXml(AWindow);
        kpmAutofill().GetDOMXml(objMessageID.value, DOMXml);
      }
    }
    catch (ErrorMessage)
    {
      kpmAutofill().MessageProcessFailed(objMessageID.value);
      kpmLog.logError('kpmAutofillEngineOnTimerFocusedWindow() Error during GetDOMXml(): ' + ErrorMessage);
    }
    
    // check if we need to get active element screen rect
    if (AllowProceed &&
        kpmAutofill().MustGetActiveElementScreenRect(objMessageID)
       )
    try
    {
      if (!AllowProceedDocument)
      {
        kpmAutofill().MessageProcessFailed(objMessageID.value);
        if (kpmDebug_LogInfo)
          kpmLog.logMessage('MustGetActiveElementScreenRect() Document process denied for <' + kpmGetDocumentUrl(ADocument) + '>');
      }
      else
      {
        AllowProceed = false;
        if (kpmDebug_LogInfo)
          kpmLog.logMessage('MustGetActiveElementScreenRect()');
        var focusedElement = kpmGetDocumentFocusedElement(ADocument);
        if (focusedElement)
        {
          var FieldType = kpmGetAdditionalFieldType(focusedElement);
          var rc = kpmGetElementScreenPosition(focusedElement);
          kpmAutofill().GetActiveElementScreenRect(objMessageID.value, rc.left, rc.top, rc.right, rc.bottom,
            (FieldType == kpm_afEdit || FieldType == kpm_afEditPassword || FieldType == kpm_afMemo));
          if (kpmDebug_GetActiveElementScreenRect)
            kpmLog.logMessage('   ActiveElementScreenRect: [' + rc.left + 'x' + rc.top + ', ' + rc.right + 'x' + rc.bottom + ']');
        } else
        {
          kpmAutofill().GetActiveElementScreenRect(objMessageID.value, 0, 0, 0, 0, false);
        }
      }
    }
    catch (ErrorMessage)
    {
      kpmAutofill().MessageProcessFailed(objMessageID.value);
      kpmLog.logError('kpmAutofillEngineOnTimerFocusedWindow() Error during GetActiveElementScreenRect(): ' + ErrorMessage);
    }
    
    // check if we need to set active element selected text
    var objSelectedText = {};
    if (AllowProceed &&
        kpmAutofill().MustSetActiveElementSelectedText(objSelectedText)
       )
    try
    {
      AllowProceed = false;
      if (kpmDebug_LogInfo)
        kpmLog.logMessage('MustSetActiveElementSelectedText()');
      var focusedElement = kpmGetDocumentFocusedElement(ADocument);
      kpmReplaceElementSelectedText(focusedElement, objSelectedText.value);
      if (kpmDebug_SetActiveElementSelectedText)
        kpmLog.logMessage('   SetActiveElementSelectedText(' + objSelectedText.value + ')');
    }
    catch (ErrorMessage)
    {
      kpmLog.logError('kpmAutofillEngineOnTimerFocusedWindow() Error during SetActiveElementSelectedText(): ' + ErrorMessage);
    }
          
    // check if we can check is active document complete
    if (AllowProceed &&
        kpmAutofill().MustSetIsActiveDocumentComplete(objMessageID)
       )
    try
    {
      if (!AllowProceedDocument)
      {
        kpmAutofill().MessageProcessFailed(objMessageID.value);
        if (kpmDebug_LogInfo)
          kpmLog.logMessage('MustSetIsActiveDocumentComplete() Document process denied for <' + kpmGetDocumentUrl(ADocument) + '>');
      }
      else
      {
        AllowProceed = false;
        if (kpmDebug_LogInfo)
          kpmLog.logMessage('MustSetIsActiveDocumentComplete()');
        var Result = kpmGetIsDocumentComplete(ADocument);
        kpmAutofill().SetIsActiveDocumentComplete(objMessageID.value, Result);
        if (kpmDebug_LogInfo)
          kpmLog.logMessage('   IsActiveDocumentComplete: ' + Result);
      }
    }
    catch (ErrorMessage)
    {
      kpmAutofill().MessageProcessFailed(objMessageID.value);
      kpmLog.logError('kpmAutofillEngineOnTimerFocusedWindow() Error during SetIsActiveDocumentComplete(): ' + ErrorMessage);
    }

    // check if we must manual autofill document
    if (AllowProceed &&
        kpmAutofill().MustManualAutofill()
       )
    try
    {
      AllowProceed = false;
      if (kpmDebug_LogInfo)
        kpmLog.logMessage('MustManualAutofill()');
      kpmAutofillDocument(AWindow, 0);
    }
    catch (ErrorMessage)
    {
      kpmLog.logError('kpmAutofillEngineOnTimerFocusedWindow() Error during ManualAutofill(): ' + ErrorMessage);
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmAutofillEngineOnTimerFocusedWindow() Error: ' + ErrorMessage);
  }
}

function kpmProcessContentMessage(ARequest)
{
  try
  {
    if (ARequest)
    {
      // spAuthKey?
      if (ARequest.name == 'spAuthKey')
      {
        // store received AuthKey
        spAuthKey = ARequest.AuthKey;
      }
      // kpmOnTabSelect?
      else if (ARequest.name == 'kpmOnTabSelect')
      {
        kpmIsActiveTab = ARequest.IsActiveTab;
        if (kpmIsActiveTab)
        {
          // access Listener from Content script
          if (kpmAutofill())
            kpmAutofill().AccessListener(true);
          kpmTabSelected(document);
        }
      }
      // kpmAutofillEngineOnTimerFocusedWindow?
      else if (ARequest.name == 'kpmAutofillEngineOnTimerFocusedWindow')
      {
        // access Listener from Content script
        if (kpmAutofill())
          kpmAutofill().AccessListener(true);
        kpmAutofillEngineOnTimerFocusedWindow();
      }
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmProcessContentMessage() Error: ' + ErrorMessage);
  }
}

// Tabs engine -----------------------------------------------------------------

var kpmIsActiveTab = true;

// This method used to check if it's active tab window
function kpmIsActiveTabWindow(AWindow)
{
  var Result = false;
  try
  {
    if (kpmIsOpera() || kpmIsChrome())
    {
      // Google Chrome, Flock3 & Opera  
      if (kpmIsActiveTab &&
          kpmAutofill() &&
          AWindow &&
          AWindow.document &&
          kpmAutofill().IsActiveTabWindow(AWindow.document.title, kpmGetWindowScreenX(AWindow), kpmGetWindowScreenY(AWindow))
         )
      {
        Result = true;
      }
    }
    else
    {
      // Firefox, Flock <= 2, Thunderbird
      Result = AWindow == kpmWindowContent(window);
    }
  }
  catch (ErrorMessage)
  {
    Result = false;
    kpmLog.logError('kpmIsActiveTabWindow() Error: ' + ErrorMessage);
  }
  return Result;
}

function kpmTabSelected(ADocument)
{
  if (kpmAutofill())
  try
  {   
    if (kpmDebug_OnTabSelect)
    {
      kpmLog.logMessage('kpmTabSelected() URL: ' + kpmGetDocumentUrl(ADocument) + 
        ', location.href: ' + ADocument.location.href + 
        ', kpmGetDocumentUrl: ' + kpmGetDocumentUrl(ADocument)
      );
    }
    if (kpmAllowProceedDocument(ADocument, false))
      kpmAutofill().AutofillMultiloginAccount(kpmGetDocumentUrl(ADocument));
    else
      kpmAutofill().HideShowedCaptionButtonBalloon();
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmTabSelected() Error: ' + ErrorMessage);
  }
}

// Web Browser engine ----------------------------------------------------------

function kpmIsOpera()
{
  try
  {
    if (opera)
      return true;
  }
  catch(ErrorMessage)
  {
    // keep silence
  }
  return false;
}

function kpmIsChrome()
{
  try
  {
    if (chrome)
      return true;
  }
  catch(ErrorMessage)
  {
    // keep silence
  }
  return false;
}

// Common unit engine  ---------------------------------------------------------

function kpmGetWindowScreenX(AWindow)
{
  var Result = 0;
  if (AWindow)
  try
  {
    if (kpmIsOpera())
      Result = AWindow.screenLeft;
    else
      Result = AWindow.screenX;
  }
  catch(ErrorMessage)
  {
    // keep silence
  }
  return Result;
}

function kpmGetWindowScreenY(AWindow)
{
  var Result = 0;
  if (AWindow)
  try
  {
    if (kpmIsOpera())
      Result = AWindow.screenTop;
    else
      Result = AWindow.screenY;
  }
  catch(ErrorMessage)
  {
    // keep silence
  }
  return Result;
}

function kpmWindowContent(AWindow)
{
  try
  {
    if (AWindow &&
        AWindow._content
       )
      return AWindow._content;
    else
      return AWindow;
  }
  catch (ErrorMessage)
  {
    return AWindow;
  }
}

function kpmGetDocumentFocusedElement(ADocument)
{
  var Result = null;
  if (ADocument)
  try
  {
    if (ADocument.commandDispatcher)
      Result = ADocument.commandDispatcher.focusedElement;
    if (!Result &&
        ADocument.commandDispatcher &&
        ADocument.commandDispatcher.focusedWindow &&
        ADocument.commandDispatcher.focusedWindow.document
       )
    {
      Result = ADocument.commandDispatcher.focusedWindow.document.activeElement;
    }
    if (!Result)
      Result = ADocument.activeElement;
    // if active element located in frame we get frame element instead real
    // acive element
    if (Result && kpmWideSameText(Result.tagName,'IFRAME') && Result.contentDocument)
    {
      Result = kpmGetDocumentFocusedElement(Result.contentDocument);        
    }       
  }    
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmGetDocumentFocusedElement() Error: ' + ErrorMessage);
  }
  return Result;
}

function kpmGetElementPosition(HtmlElement)
{
  if (HtmlElement == null)
    return {'left': 0, 'top': 0, 'right': 0, 'bottom': 0};
  var left = 0;
  var top = 0;
  var width = 0;
  var height = 0;
  if (HtmlElement.getBoundingClientRect)
  {
    // from version 3 & heigher
    var rc = HtmlElement.getBoundingClientRect();
    left = rc.left;
    top = rc.top;
    width = rc.right - rc.left;
    height = rc.bottom - rc.top;
  } else
  {
    var parentNode = null;
    var offsetParent = null;    
    offsetParent = HtmlElement.offsetParent;
    var originalElement = HtmlElement;
    var Element = HtmlElement; // Element will be nodes as we walk up, HtmlElement will be saved for offsetParent references
    while (Element.parentNode != null)
    {
      Element = Element.parentNode;
      // calc scrollTop
      if (Element.scrollTop && Element.scrollTop > 0)
        top -= Element.scrollTop;
      if (Element.scrollLeft && Element.scrollLeft > 0)
        left -= Element.scrollLeft;
      // if this node is also the offsetParent, add on the offsets and reset to the new offsetParent
      if (Element == offsetParent)
      {
        left += HtmlElement.offsetLeft;
        top += HtmlElement.offsetTop;
        if (Element.nodeName != 'TABLE')
        {
          if (Element.clientLeft)
            left += Element.clientLeft;
          if (Element.clientTop)
            top += Element.clientTop;
        }
        HtmlElement = Element;
        if (HtmlElement.offsetParent == null)
        {
          if (HtmlElement.offsetLeft)
            left += HtmlElement.offsetLeft;
          if (HtmlElement.offsetTop)
            top += HtmlElement.offsetTop;
        }
        offsetParent = HtmlElement.offsetParent;
      }
    }
    if (originalElement.offsetWidth)
      width = originalElement.offsetWidth;
    if (originalElement.offsetHeight)
      height = originalElement.offsetHeight;
  }
  return {'left': left, 'top': top, 'right': left+width, 'bottom': top+height};
}

function kpmGetElementScreenPosition(HtmlElement)
{
  // obtain element position relative to document
  var rc = kpmGetElementPosition(HtmlElement);
  // check for kpmAutofill object initialized
  if (kpmAutofill())
  {
    // obtain document position relative to screen
    var ADocument = HtmlElement.ownerDocument;
    // calc element's screen rect
    //   1. try to use Mozilla properties
    //      this applied to Mozilla Firefox, Flock 1.2, SeaMonkey
    if (ADocument && ADocument.defaultView)
    try
    {
      var AWindow = ADocument.defaultView;
      var x = AWindow.mozInnerScreenX;
      var y = AWindow.mozInnerScreenY;
      if ((typeof x != 'undefined') &&
          (typeof y != 'undefined') &&
          (x || y)
         )
      {
        var domWindowUtils = AWindow.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
          .getInterface(Components.interfaces.nsIDOMWindowUtils);
        if (domWindowUtils && domWindowUtils.screenPixelsPerCSSPixel)
        {
          rc.left = parseInt((x + rc.left) * domWindowUtils.screenPixelsPerCSSPixel);
          rc.top = parseInt((y + rc.top) * domWindowUtils.screenPixelsPerCSSPixel);
          rc.right = parseInt((x + rc.right) * domWindowUtils.screenPixelsPerCSSPixel);
          rc.bottom = parseInt((y + rc.bottom) * domWindowUtils.screenPixelsPerCSSPixel);
          return rc;
        }
      }
    }
    catch (ErrorMessage)
    {
      // keep silence
    }	
	// Opera and Chrome returns wrong coordinates of element located in iframe 
	if (kpmIsChrome() || kpmIsOpera())
	try
    {
	  var AWindow = ADocument.defaultView;	 
	  // In Chrome AWindow == null. In Opera we check AWindow.frameElement
	  if (AWindow == null || AWindow.frameElement)
	  {
	    var frames = window.document.getElementsByTagName('iframe');		 
	    for(var i = 0; i < frames.length; i++)
	    {		
	      var tmpFrame = frames[i];
	      if (ADocument == tmpFrame.contentDocument)
	      {			
	        var frameRC = kpmGetElementPosition(tmpFrame);
	        rc.left = rc.left + frameRC.left;
	        rc.top = rc.top + frameRC.top;
	        rc.right = rc.right + frameRC.right;
	        rc.bottom = rc.bottom + frameRC.bottom;
			break;
	      }
	    }
	  }
	}
    catch (ErrorMessage)
    {
      // keep silence
    }	
	
	
    //   2. try to use CalcElementScreenRect()
    //      this applied to Google Chrome, Flock 3, Opera
    var left = {'value': rc.left};
    var top = {'value': rc.top};
    var right = {'value': rc.right};
    var bottom = {'value': rc.bottom};
    if (kpmAutofill().CalcElementScreenRect(kpmWindowContent(window).document.title, left, top, right, bottom))
    {
      rc.left = left.value;
      rc.top = top.value;
      rc.right = right.value;
      rc.bottom = bottom.value;
    }
  }
  return rc;
}

function kpmGetElementName(HtmlElement)
{
  var Result = '';
  if (HtmlElement)
  {
    Result = HtmlElement.name;
    // Opera bugfix - some forms returns their names as an input element.
    if (Result && typeof Result != 'string')
      Result = HtmlElement.getAttribute('name');
  }
  return Result;
}
 
function kpmReplaceElementSelectedText(HtmlElement, replaceString)
{
  var FieldType = kpmGetAdditionalFieldType(HtmlElement);
  if (((FieldType == kpm_afEdit) ||
       (FieldType == kpm_afEditPassword) ||
       (FieldType == kpm_afMemo)
      ) &&
      HtmlElement.setSelectionRange
     ) 
  {
    var selectionStart = HtmlElement.selectionStart;
    var selectionEnd = HtmlElement.selectionEnd;
    HtmlElement.value = HtmlElement.value.substring(0, selectionStart) + replaceString + HtmlElement.value.substring(selectionEnd);
    var pos = selectionStart + replaceString.length;
    HtmlElement.setSelectionRange(pos, pos);
    kpmHTMLElementSetFocus(HtmlElement);
  }
}

function kpmRecurseGetHTMLDocumentListFromFrames(AFrames, AList)
{
  if (AFrames && AList)
  try
  {
    for(var i = 0; i < AFrames.length; i++)
      kpmRecurseGetHTMLDocumentList(AFrames[i], AList);
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmRecurseGetHTMLDocumentListFromFrames() Error: ' + ErrorMessage);
  }
}
  
function kpmRecurseGetHTMLDocumentList(AWindow, AList)
{
  if (AWindow && AList)
  try
  {
    // 1. add Window's document
    //   1.1. obtain the document
    var ADocument = null;
    if (AWindow.contentDocument)
      ADocument = AWindow.contentDocument; // it is "iframe" element
    else
      ADocument = AWindow.document;        // it is "window" element
    if (!ADocument)
      return;
    //   1.2. check if we can access the document's properties (cross-domain access)?
    try
    {
      // check if we can access the document?
      var t = ADocument.length && ADocument.forms;
    }
    catch (ErrorMessage)
    {
      // cross-domain error - just ignore specified window/frame
      return;
    }
    //   1.3. check if the document has been already added
    for(var i = 0; i < AList.length; i++)
    {
      var d = AList[i];
      if (d == ADocument)
        return;
    }
    //   1.4. add the document to the list
    AList[AList.length] = ADocument;
    
    // 2. try to recurse the Window's frames
    kpmRecurseGetHTMLDocumentListFromFrames(AWindow.frames, AList);
    
    // 3. try to recurse the Document's iframes
    //      INFO: This is made only for Google Chrome because "window.frames[i]" 
    //      returns "null" for "iframe" element (is it a bug or security?).
    if (kpmIsChrome())
    {
      var frames = null;
      // 3.1. enum <iframe> elements
      frames = ADocument.getElementsByTagName('iframe');
      kpmRecurseGetHTMLDocumentListFromFrames(frames, AList);
      // 3.2. enum <frame> elements
      frames = ADocument.getElementsByTagName('frame');
      kpmRecurseGetHTMLDocumentListFromFrames(frames, AList);
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmRecurseGetHTMLDocumentList() Error: ' + ErrorMessage);
  }
}

function kpmIsHTMLDocument(ADocument)
{
  if (!ADocument)
    return false;
  if (kpmIsOpera())
  {
    try
    {
      // WARNING: Filter child frames! 'window.parent = window' - the window parent equal itself if there is no parent
      var AWindow = ADocument.defaultView;
      // check for frame element
      if (AWindow.frameElement || AWindow != AWindow.parent)
        return false;
      return true;
    }
    catch (ErrorMessage)
    {
      return false;
    }
  }
  else
  {
    return ADocument instanceof HTMLDocument;
  }
}

function kpmGetHTMLDocumentList(AWindow)
{
  var Result = new Array();
  kpmRecurseGetHTMLDocumentList(AWindow, Result);
  return Result;
}

function kpmRecurseGetHTMLElementList(AElements, AList, AAllowKnownTypes, 
  ANeedInputElementType, AMustSortList, ALowSourceIndex, AHighSourceIndex)
{
  if (AElements && AList)
  try
  {
    for(var i = 0; i < AElements.length; i++)
    {
      var Element = AElements[i];
      // filter by element type
      if (!kpmWideIsEmpty(ANeedInputElementType) &&
          !kpmWideSameText(ANeedInputElementType, Element.type)
         )
      {
        // found unneed element type
        continue;
      }
      var FieldType = kpmGetAdditionalFieldType(Element);
      // filter by SourceIndex
      if (ALowSourceIndex > -1 && AHighSourceIndex > -1)
      {
        var SourceIndex = kpmGetHTMLElementSourceIndex(Element);
        if (SourceIndex && SourceIndex >= ALowSourceIndex && SourceIndex <= AHighSourceIndex)
        {
          // supported element
        } else
        {
          // unsupported element
          continue;
        }
      }
      // try to add element to list
      if (AAllowKnownTypes)
      {
        if (FieldType != kpm_afUnknown &&
            AList.indexOf(Element) == -1
           )
        {
          AList[AList.length] = Element;
          AMustSortList.value = true;
        }
      } else
      {
        if (FieldType == kpm_afButtonSubmit &&
            AList.indexOf(Element) == -1
           )
        {
          AList[AList.length] = Element;
          AMustSortList.value = true;
        }
      }
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmRecurseGetHTMLElementList() Error: ' + ErrorMessage);
  }
}

function kpmGetIsDocumentComplete(ADocument)
{
  var Result = false;
  if (ADocument && ADocument.body)
    Result = !kpmWideIsEmpty(ADocument.body.getAttribute('kpmIsDocumentComplete'));
  return Result;
}

function kpmSetIsDocumentComplete(ADocument)
{
  if (ADocument && ADocument.body)
    ADocument.body.setAttribute('kpmIsDocumentComplete', '666');
}

function kpmHTMLElementListCompare(Element1, Element2)
{
  var Index1 = kpmGetHTMLElementSourceIndex(Element1);
  var Index2 = kpmGetHTMLElementSourceIndex(Element2);
  if (Index1 && Index2)
    return Index1 - Index2;
  else
    return 0;
}

function kpmGetHTMLElementSourceIndex(AElement)
{
  var Result;
  if (AElement)
  {
    Result = AElement.getAttribute('sourceIndex');
    if (!Result)
    {
      var Document = AElement.ownerDocument;
      if (Document)
      {
        if (!Document.body.getAttribute('kpmSourceIndexSet'))
        {
          Document.body.setAttribute('kpmSourceIndexSet', '666');
          var all = Document.all || Document.getElementsByTagName('*');
          if (all)
          {
            // INFO: We need to backup the all elements list, cause some other addons can dynamically
            //   add any elements to the document and our enum cycle can hang the browser!
            //   This bug reproducable on http://hattrick.org/ with 'FoxTrick' add-on.
            // 1. backup elements list
            var len = all.length;
            var list = new Array(len);
            for(var k = len-1; k >= 0; k--)
              list[k] = all[k];
            all = list;
            // 2. enum elements
            var SourceIndex = 0;
            for(var k = 0; k < all.length; k++)
            {
              var Element = all[k];
              if (kpmGetAdditionalFieldType(Element) != kpm_afUnknown)
              {
                Element.setAttribute('sourceIndex', SourceIndex);
                if (Element == AElement)
                  Result = SourceIndex;
                SourceIndex = SourceIndex + 1;
              }
            }
            // 3. clear elements list
            all.length = 0;
          }
        }
      }
    }
  }
  return Result;
}
    
function kpmGetHTMLElementList(AForm, ADocumentsInputElements)
{
  const MaxElementsCount = 125;
  const MaxDocumentsInputElementsCount = 125;
  var Result = new Array();
  if (AForm)
  try
  {
    var Elements;
    var MustSortList = {};
    // 1. add Form.item() elements
    var addedElements = 0;
    for(var i = 0; i < AForm.elements.length; i++)
    {
      var Element = AForm.elements[i];
      // fix for www.hostmonster.com which has lot of checkbox elements on identity form
      if (kpmCompareIDorName(Element, 'addon_domain') && kpmWideSearchText('hostmonster.com', AForm.action))
      {
        continue;
      } 
      Result[Result.length] = Element;
      addedElements++;
      if (addedElements > MaxElementsCount)
        break;
    }
    // WARNING: Deny to return more than 75 elements
    if (AForm.elements.length <= MaxElementsCount)
    {
      // 2. add additional elements of AForm
      //   2.1. add <input type="image"> elements
      Elements = AForm.getElementsByTagName('input');
      kpmRecurseGetHTMLElementList(Elements, Result, true, kpm_szIMAGE, MustSortList, -1, -1);
      //   2.2. add <a> elements
      Elements = AForm.getElementsByTagName('a');
      kpmRecurseGetHTMLElementList(Elements, Result, false, '', MustSortList, -1, -1);
      //   2.3. add <img> elements
      Elements = AForm.getElementsByTagName('img');
      kpmRecurseGetHTMLElementList(Elements, Result, false, '', MustSortList, -1, -1);
      // 3. add additional elements of AForm.parentNode = <table>|<tbody>
      if (AForm.parentNode &&
          (kpmWideSameText(AForm.parentNode.tagName, 'table') ||
           kpmWideSameText(AForm.parentNode.tagName, 'tbody')
          )
         )
      {
        var parentElement = AForm.parentNode;
        //   3.1. add <input type="image"> elements
        Elements = parentElement.getElementsByTagName('input');
        kpmRecurseGetHTMLElementList(Elements, Result, true, kpm_szIMAGE, MustSortList, -1, -1);
        //   3.2. add <a> elements
        Elements = parentElement.getElementsByTagName('a');
        kpmRecurseGetHTMLElementList(Elements, Result, false, '', MustSortList, -1, -1);
        //   3.3. add <img> elements
        Elements = parentElement.getElementsByTagName('img');
        kpmRecurseGetHTMLElementList(Elements, Result, false, '', MustSortList, -1, -1);
      }
      // 4. add additional elements of ADocument that lays after within AForm & next one
      var ADocument = AForm.ownerDocument;
      if (ADocument && ADocumentsInputElements && ADocumentsInputElements.length <= MaxDocumentsInputElementsCount)
      {
        // obtain indexes of AForm & next after it
        var LowIndex = -1;
        var HighIndex = -1;
        for(var i = 0; i < ADocument.forms.length; i++)
        {
          if (ADocument.forms[i] == AForm)
          {
            LowIndex = i;
            if (i < ADocument.forms.length-1)
              HighIndex = i+1;
            break;
          }
        }
        if (LowIndex != -1)
        {
          // obtain source indexes of forms
          LowIndex = parseInt(kpmGetHTMLElementSourceIndex(AForm));
          if (HighIndex == -1)
            HighIndex = 999999;
          else
            HighIndex = parseInt(kpmGetHTMLElementSourceIndex(ADocument.forms[HighIndex]));
          if (LowIndex != HighIndex && HighIndex - LowIndex > 1)
          {
            LowIndex = LowIndex + 1;
            HighIndex = HighIndex - 1;
            // 4.1. add <input type="image"> elements          
            kpmRecurseGetHTMLElementList(ADocumentsInputElements, Result, true, kpm_szIMAGE, MustSortList, LowIndex, HighIndex);
          }
        }
      }
    }
    // sort the list if need
    if (MustSortList.value && Result.length > 1)
      Result.sort(kpmHTMLElementListCompare);
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmGetHTMLElementList() Error: ' + ErrorMessage);
  }
  return Result;
}

function kpmGetHTMLDocumentElementList(ADocument)
{
  var Result = new Array();
  // don't export for Document if any form exists in document!
  if (ADocument/* &&
      (!ADocument.forms ||
       !ADocument.forms.length
      )*/
     )
  try
  {
    var Elements;
    var MustSortList = {};
    // 1.1. add <input> elements
    Elements = ADocument.getElementsByTagName('input');
    kpmRecurseGetHTMLElementList(Elements, Result, true, '', MustSortList);
    // 1.2. add <select> elements
    Elements = ADocument.getElementsByTagName('select');
    kpmRecurseGetHTMLElementList(Elements, Result, true, '', MustSortList);
    // 1.3. add <a> elements
    Elements = ADocument.getElementsByTagName('a');
    kpmRecurseGetHTMLElementList(Elements, Result, false, '', MustSortList);
    // 1.4. add <img> elements
    Elements = ADocument.getElementsByTagName('img');
    kpmRecurseGetHTMLElementList(Elements, Result, false, '', MustSortList);
    // sort the list if need
    if (MustSortList.value && Result.length > 1)
      Result.sort(kpmHTMLElementListCompare);
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmGetHTMLDocumentElementList() Error: ' + ErrorMessage);
  }
  return Result;
}

function kpmGetDocumentUrl(ADocument)
{
  var Result = '';
  if (ADocument)
  {
    try
    {
      netscape.security.PrivilegeManager.enablePrivilege('UniversalBrowserRead');
    }
    catch (ErrorMessage)
    {
      // keep silence
    }
    try
    {
      if (ADocument.location)
        Result = ADocument.location.href;
      else
        Result = ADocument.URL;
    }
    catch (ErrorMessage)
    {
      Result = ADocument.URL;
    }
  }
  return Result;
}

function kpmIsHTMLElementVisible(HtmlElement)
{
  if (HtmlElement)
  try
  {
    var ADocument = HtmlElement.ownerDocument;
    if (ADocument == null)
      ADocument = document;
    if (ADocument == null)
      return true;
    if (kpmDebug_IsHTMLElementVisible_Details)
      kpmLog.logMessage('kpmIsHTMLElementVisible()');
    if (kpmWideSearchText('playboy.com/', kpmGetDocumentUrl(ADocument)) ||
        kpmWideSearchText('icq.com/', kpmGetDocumentUrl(ADocument)) ||
        kpmWideSearchText('alfabank.ru/', kpmGetDocumentUrl(ADocument)) ||
        kpmWideSearchText('comcast.net/', kpmGetDocumentUrl(ADocument))
       )
    {
      return true;
    }
    var Element = HtmlElement;
    while (Element)
    {
      if (Element == ADocument || kpmWideSameText(Element.tagName, 'body'))
        return true;
      var OverflowHidden = false;
      // check style
      if (kpmDebug_IsHTMLElementVisible_Details)
        kpmLog.logMessage('  Element <' + Element.tagName + ':' + kpmGetElementName(Element) + '/' +  Element.id + '> ' +
          'value: <' + kpmGetAdditionalFieldValue(Element) + '>'
        );
      if (Element.style)
      {
        if (kpmDebug_IsHTMLElementVisible_Details)
          kpmLog.logMessage(
            '    Element.style.display: <' + Element.style.display + '> ' +
            '    Element.style.visibility: <' + Element.style.visibility + '>'
          );
        // check style.display value
        if (kpmWideSameText(Element.style.display, 'none'))
        {
          if (kpmDebug_IsHTMLElementVisible)
            kpmLog.logMessage('kpmIsHTMLElementVisible() <' + Element.tagName + ':' + kpmGetElementName(Element) + '/' +  Element.id + '> ' +
              'value: <' + kpmGetAdditionalFieldValue(Element) + '> ' +
              'Element.style.display: ' + Element.style.display);
          return false;
        }
        // check style.visibility value
        if (kpmWideSameText(Element.style.visibility, 'hidden'))
        {
          if (kpmDebug_IsHTMLElementVisible)
            kpmLog.logMessage('kpmIsHTMLElementVisible() <' + Element.tagName + ':' + kpmGetElementName(Element) + '/' +  Element.id + '> ' +
              'value: <' + kpmGetAdditionalFieldValue(Element) + '> ' +
              'Element.style.visibility: ' + Element.style.visibility);
          return false;
        }
        // check style.overflow value
        if (kpmWideSameText(Element.style.overflow, 'hidden'))
          OverflowHidden = true;
      }
      // form can be located after <table>/<tr>/<body> tags 
      // and by such method it can be hidden (bug of IE/Firefox?)
      var FieldType = kpmGetAdditionalFieldType(Element);     
      if (FieldType == kpm_afHtmlForm &&
          Element.parentNode &&
          (kpmWideSameText(Element.parentNode.tagName, 'table') ||
           kpmWideSameText(Element.parentNode.tagName, 'tbody') ||
           kpmWideSameText(Element.parentNode.tagName, 'tr') ||
           kpmWideSameText(Element.parentNode.tagName, 'body')
          )
         )
      {
        // ignore currentStyle checking!
        if (kpmDebug_IsHTMLElementVisible)
          kpmLog.logMessage('kpmIsHTMLElementVisible() <' + Element.tagName + ':' + kpmGetElementName(Element) + '/' +  Element.id + '> ' +
            'value: <' + kpmGetAdditionalFieldValue(Element) + '> ' +
            'form located after <table>/<tbody>/<tr>/<body> tag!');
      } else
      {
        // check currentStyle
        if (ADocument.defaultView && ADocument.defaultView.getComputedStyle)
        try
        {
          var computedStyle = ADocument.defaultView.getComputedStyle(Element, '');
          if (computedStyle)
          {
            if (kpmDebug_IsHTMLElementVisible_Details)
              kpmLog.logMessage(
                '    Element.computedStyle.display: <' + computedStyle.display + '> ' +
                '    Element.computedStyle.visibility: <' + computedStyle.visibility + '>'
              );
            // check currentStyle.display value
            if (kpmWideSameText(computedStyle.display, 'none'))
            {
              if (kpmDebug_IsHTMLElementVisible)
                kpmLog.logMessage('kpmIsHTMLElementVisible() <' + Element.tagName + ':' + kpmGetElementName(Element) + '/' +  Element.id + '> ' +
                  'value: <' + kpmGetAdditionalFieldValue(Element) + '> ' +
                  'computedStyle.display: ' + computedStyle.display);
              return false;
            }              
            // check currentStyle.visibility value
            if (kpmWideSameText(computedStyle.visibility, 'hidden'))
            {
              if (kpmDebug_IsHTMLElementVisible)
                kpmLog.logMessage('kpmIsHTMLElementVisible() <' + Element.tagName + ':' + kpmGetElementName(Element) + '/' +  Element.id + '> ' +
                  'value: <' + kpmGetAdditionalFieldValue(Element) + '> ' +
                  'computedStyle.visibility: ' + computedStyle.visibility);
              return false;
            }
            // check currentStyle.overflow value
            if (kpmWideSameText(computedStyle.overflow, 'hidden'))
              OverflowHidden = true;
          }
        }
        catch (ErrorMessage)
        {
          kpmLog.logError('kpmIsHTMLElementVisible() GetComputedStyle() error: ' + ErrorMessage);
        }
        // check clientWidth & clientHeight values
        if (!Element.clientWidth || !Element.clientHeight)
        {
          var rc = kpmGetElementPosition(Element);
          var width = rc.right - rc.left;
          var height = rc.bottom - rc.top;
          if (OverflowHidden &&
              ((!Element.clientWidth && !width) ||
               (!Element.clientHeight && !height)
              )
             )
          {
            if (kpmDebug_IsHTMLElementVisible)
              kpmLog.logMessage('kpmIsHTMLElementVisible() <' + Element.tagName + ':' + kpmGetElementName(Element) + '/' +  Element.id + '> ' +
                'value: <' + kpmGetAdditionalFieldValue(Element) + '> ' +
                'Element.(computed)Style.overflow: hidden, ' +
                'Element.clientWidth: ' + Element.clientWidth + ', ' +
                'Element.clientHeight: ' + Element.clientHeight + ' ' + 
                'kpmGetElementPosition(width, height): ' + width + 'x' + height
              );
            return false;
          }
        }
      }
      // goto to parent node/element
      Element = Element.parentNode; // ?? Element.parentElement;
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmIsHTMLElementVisible() Error: ' + ErrorMessage);
  }
  return true;
}

function kpmIsHTMLElementReadOnly(HtmlElement)
{
  var Result = false;
  if (HtmlElement)
  try
  {
    Result = HtmlElement.readOnly;
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmIsHTMLElementReadOnly() Error: ' + ErrorMessage);
  }
  return Result;
}

function kpmIsHTMLElementDisabled(HtmlElement)
{
  var Result = false;
  if (HtmlElement)
  try
  {
    Result = HtmlElement.disabled;
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmIsHTMLElementDisabled() Error: ' + ErrorMessage);
  }
  return Result;
}

function kpmHTMLElementSetFocus(HtmlElement)
{
  if (!HtmlElement)
    return false;
  try
  {
    HtmlElement.focus();
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmHTMLElementSetFocus() Error: ' + ErrorMessage);
  }
  return true;
}

function kpmHTMLElementClick(HtmlElement)
{
  if (!HtmlElement)
    return false;
  try
  {
    if (HtmlElement.click)
      HtmlElement.click(); 
    else
      kpmHTMLElementFireMouseEvent(HtmlElement, 'click');
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmHTMLElementClick() Error: ' + ErrorMessage);
  }
  return true;
}

function kpmHTMLElementFireEvent(HtmlElement, EventName)
{
  if (!HtmlElement || !EventName)
    return false;
  try
  {
    var event = HtmlElement.ownerDocument.createEvent('HTMLEvents');
    if (event)
    {
      event.initEvent(EventName, 
        true,  // canBubble
        true   // cancelable
      );
      HtmlElement.dispatchEvent(event);
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmHTMLElementFireEvent() Error: ' + ErrorMessage);
  }
  return true;
}

function kpmHTMLElementFireKeyEvent(HtmlElement, EventName, KeyCode)
{
  if (!HtmlElement || !EventName)
    return false;
  try
  {
    var event = HtmlElement.ownerDocument.createEvent('KeyboardEvent');
    if (event)
    {
      if (event.initKeyEvent)
      {
        // Gecko declaration
        event.initKeyEvent(EventName,
          true,       // canBubble
          true,       // cancelable
          HtmlElement.ownerDocument.defaultView, // view
          false,      // ctrlKey
          false,      // altKey
          false,      // shiftKey
          false,      // metaKey
          KeyCode,    // keyCode
          0           // charCode
        );
      }
      else
      {
        // DOM3 declaration
        event.initKeyboardEvent(EventName,
          true,       // canBubble
          true,       // cancelable
          HtmlElement.ownerDocument.defaultView, // view
          false,      // ctrlKey
          false,      // altKey
          false,      // shiftKey
          false,      // metaKey
          KeyCode,    // keyCode
          0           // charCode
        );
      }
      HtmlElement.dispatchEvent(event);
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmHTMLElementFireKeyEvent() Error: ' + ErrorMessage);
  }
  return true;
}

function kpmHTMLElementFireMouseEvent(HtmlElement, EventName)
{
  if (!HtmlElement || !EventName)
    return false;
  try
  {
    var event = HtmlElement.ownerDocument.createEvent('MouseEvents');
    if (event)
    {
      event.initMouseEvent(EventName, 
        true,  // canBubble
        true,  // cancelable
        HtmlElement.ownerDocument.defaultView, // view 
        1,     // mouse click count - i.e. once
        0,     // screenX
        0,     // screenY
        0,     // clientX
        0,     // clientY
        false, // ctrlKey
        false, // altKey
        false, // shiftKey
        false, // metaKey
        0,     // button
        null   // relatedTarget
      );
      HtmlElement.dispatchEvent(event);
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmHTMLElementFireMouseEvent() Error: ' + ErrorMessage);
  }
  return true;
}

function kpmHTMLFormSubmit(HtmlFormElement, Force)
{
  var Result = false;
  try
  {
    Result =
      (HtmlFormElement != null) &&
      (Force ||
       (HtmlFormElement.action != '')
      );
    if (Result)
    {
      // call form.onsubmit() event
      try
      {
        if (HtmlFormElement.fireEvent)
        {
          // Internet Explorer
          if (!HtmlFormElement.fireEvent('onsubmit', document.createEventObject()))
            return Result;
        } else
        if (document.createEvent && HtmlFormElement.dispatchEvent)
        {
          // Gecko
          var event = document.createEvent('Events');
          event.initEvent('submit', false, true);
          if (!HtmlFormElement.dispatchEvent(event))
            return Result;
        } else
        if (HtmlFormElement.onsubmit)
        {
          // manual onsubmit() call
          if (HtmlFormElement.onsubmit() == false)
            return Result;
        }        
      }
      catch (ErrorMessage)
      {
        // keep silence
      }
      // submit form
      HtmlFormElement.submit();
    }
  }
  catch (ErrorMessage)
  {
    Result = false;
    kpmLog.logError('kpmHTMLFormSubmit() Error: ' + ErrorMessage);
  }
  return Result;
}

function kpmGetHTMLFormMethod(HtmlFormElement)
{
  if (!HtmlFormElement)
    return kpm_htmlFormMethod_Unknown; else
  if (kpmWideSameText(HtmlFormElement.method, 'post'))
    return kpm_htmlFormMethod_Post; else
  {
    /* On http://rutube.ru/ & http://tv-rip.ru/ form has get method,
         but it doesn't submitting, i.e. <form onsubmit="...; return false;">
         in such case let's think that form has post method!
    */
    var attributeName = 'onsubmit';
    var strAttribute = '';
    var funAttribute = HtmlFormElement.getAttribute(attributeName);
    if (funAttribute)
      strAttribute = funAttribute.toString();
    if (kpmWideSearchText('return false', strAttribute))
      return kpm_htmlFormMethod_Post;
    else
      return kpm_htmlFormMethod_Get;
  }
}

// TkpmXMLSerializer ------------------------------------------------------------

function kpmNewXMLParser()
{
  if (typeof DOMParser != 'undefined')
    return new DOMParser();
  else
    return new window.DOMParser();
}

function kpmNewXMLSerializer()
{
  if (typeof XMLSerializer != 'undefined')
    return new XMLSerializer();
  else
    return new TkpmXMLSerializer();
}
        
function TkpmXMLSerializer()
{
  this.Result = '';
}

TkpmXMLSerializer.prototype =
{
  serializeXmlNodeToString: function(XmlNode)
  {
    if (!XmlNode)
      return;
    this.Result = this.Result + '<' + XmlNode.nodeName;
    // add attributes (if exists)
    if (XmlNode.hasAttributes)
    {
      for(var i = 0; i < XmlNode.attributes.length; i++)
      {
        var attribute = XmlNode.attributes[i];
        this.Result = this.Result + ' ' + attribute.name + '="' + attribute.value + '"';
      }
    }
    // add child nodes (if exists)
    if (XmlNode.hasChildNodes())
    {
      this.Result = this.Result + '>';
      for(var i = 0; i < XmlNode.childNodes.length; i++)
      {
        var childXmlNode = XmlNode.childNodes[i];
        this.serializeXmlNodeToString(childXmlNode);
      }
      this.Result = this.Result + '</' + XmlNode.nodeName + '>';
    }
    else
    {
      this.Result = this.Result + '/>';
    }
  },
  
  serializeToString: function(XmlDocument)
  {
    this.Result = '';
    if (XmlDocument)
      this.serializeXmlNodeToString(XmlDocument.documentElement);
    return this.Result;
  }
}

// TkpmPasswordControlLastValue -------------------------------------------------

function TkpmPasswordControlLastValue()
{
  this.EditPassword = null;
  this.value = '';
}

TkpmPasswordControlLastValue.prototype =
{
  // WARNING: We mustn't declare here variables, cause when pass this object
  //   into function - variables would be reset to default values!

  clear: function()
  {
    this.EditPassword = null;
    this.value = '';
  },

  get: function(AEditPassword)
  {
    if (this.EditPassword == AEditPassword)
      return this.value
    else
      return '';
  },
  
  set: function(AEditPassword)
  {
    this.EditPassword = AEditPassword;
    if (AEditPassword)
      this.value = AEditPassword.value;
    else
      this.value = '';
  },
  
  isEmpty: function(AEditPassword)
  {
    return kpmWideIsEmpty(this.get(AEditPassword));
  }
}

var kpmPasswordControlLastValue = new TkpmPasswordControlLastValue();

// TkpmPasswordControlExistsInfo ------------------------------------------------

function TkpmPasswordControlExistsInfo()
{
  this.Password = {};
  this.PasswordControlRect = {'left': 0, 'top': 0, 'right': 0, 'bottom': 0};
  this.FormAction = {};
  this.FormMethod = kpm_htmlFormMethod_Unknown;
  this.PasswordControlEmpty = true;
  this.PasswordControlMustBeFilled = false;
  this.MultipageAuthorizationLogin = {};
  this.MultipageAuthorizationLoginControlRect = {'left': 0, 'top': 0, 'right': 0, 'bottom': 0};
}

TkpmPasswordControlExistsInfo.prototype =
{
  // WARNING: We mustn't declare here variables, cause when pass this object
  //   into function - variables would be reset to default values!
}

// TkpmAutofillInfo -------------------------------------------------------------

function TkpmAutofillInfo()
{
  this.Login = {};
  this.RealLogin = {};
  this.Password = {};
  this.AutoSubmit = false;
  this.AllowSetFocus = false;
  this.LoginAutofilled = false;
  this.PasswordAutofilled = false;
  this.IdentityFormIndex = 666;
  this.AutofillReason = 0;
  this.iForm = -1;
}

TkpmAutofillInfo.prototype =
{
  // WARNING: We mustn't declare here variables, cause when pass this object
  //   into function - variables would be reset to default values!
}

// TkpmAdditionalFieldsAutofillInfo ---------------------------------------------

function TkpmAdditionalFieldsAutofillInfo()
{
  this.SubmitFormElement = null;
  this.SubmitButtonElement = null;
  this.CheckAvailabilityButtonElement = null;
  this.FocusingElement = null;
  this.FocusingButtonElement = null;
  this.FocusingButtonImageElement = null;
  this.FocusingButtonSubmitElement = null;
  this.MustFindFocusingButtonElement = 0;
}

TkpmAdditionalFieldsAutofillInfo.prototype =
{
  // WARNING: We mustn't declare here variables, cause when pass this object
  //   into function - variables would be reset to default values!
}

// Additional Fields engine ----------------------------------------------------

function kpmIsInputAdditionalFieldTypes(FieldType)
{
  if ((FieldType == kpm_afButton) ||
      (FieldType == kpm_afButtonImage) ||
      (FieldType == kpm_afButtonSubmit) ||
      (FieldType == kpm_afEdit) ||
      (FieldType == kpm_afEditPassword) ||
      (FieldType == kpm_afCheckBox) ||
      (FieldType == kpm_afRadioButton)
     )
    return true;
  else
    return false;
}

function kpmIsTextAreaAdditionalFieldTypes(FieldType)
{
  if ((FieldType == kpm_afMemo))
    return true;
  else
    return false;
}

function kpmIsSelectAdditionalFieldTypes(FieldType)
{
  if ((FieldType == kpm_afComboBox) ||
      (FieldType == kpm_afComboBoxDropDownList) ||
      (FieldType == kpm_afListBox)
     )
    return true;
  else
    return false;
}

function kpmAdditionalFieldTypeToStr(FieldType)
{
  if (FieldType == kpm_afUnknown)
    return 'afUnknown'; else
  if (FieldType == kpm_afHtmlForm)
    return 'afHtmlForm'; else
  if (FieldType == kpm_afButton)
    return 'afButton'; else
  if (FieldType == kpm_afButtonImage)
    return 'afButtonImage'; else
  if (FieldType == kpm_afButtonSubmit)
    return 'afButtonSubmit'; else
  if (FieldType == kpm_afEdit)
    return 'afEdit'; else
  if (FieldType == kpm_afEditPassword)
    return 'afEditPassword'; else
  if (FieldType == kpm_afCheckBox)
    return 'afCheckBox'; else
  if (FieldType == kpm_afRadioButton)
    return 'afRadioButton'; else
  if (FieldType == kpm_afMemo)
    return 'afMemo'; else
  if (FieldType == kpm_afComboBox)
    return 'afComboBox'; else
  if (FieldType == kpm_afComboBoxDropDownList)
    return 'afComboBoxDropDownList'; else
  if (FieldType == kpm_afListBox)
    return 'afListBox';
  else
    return 'WARNING: Undefined FieldType!';
}

var kpmRealAdditionalFieldType = false;

function kpmIsButtonSubmitIDorName(Element)
{
  var Result = false;
  if (kpmCompareIDorName(Element, '_Login1_Login') ||         // postani-student.hr
      kpmCompareIDorName(Element, 'SignIn') ||
      kpmCompareIDorName(Element, 'SignOn') ||
      kpmCompareIDorName(Element, 'btn_continue') ||
      kpmCompareIDorName(Element, 'btnDoLogin') ||
      kpmCompareIDorName(Element, 'btnLogin') ||
      kpmCompareIDorName(Element, 'sibLogOn') ||         // trading.scottrade.com 
      kpmCompareIDorName(Element, 'authcode_Button') ||  // klient2.rb.cz
      kpmCompareIDorName(Element, 'iknoplogin') || 
      (kpmCompareIDorName(Element, 'Login') &&  
       kpmCompareIDorName(Element, '_btn')                     
      ) ||       
      kpmCompareIDorName(Element, 'loginSubmit')	||			// me.memeo.com
	    kpmCompareIDorName(Element, 'login_submit')		    // weibo.com
     )
  {
    Result = true;
  }
  if (Result &&
	  (kpmCompareIDorName(Element, 'forgot') ||
	   kpmCompareIDorName(Element, 'forget')
	  )
	 )
  {
	Result = false;
  }
  return Result;
}

function kpmIsButtonSubmitAttributeValue(AttributeValue, Url)
{
  var Result = false;
  if (kpmWideSearchText('sign in', AttributeValue) ||
      kpmWideSearchText('sign&nbsp;in', AttributeValue) ||
      (kpmWideSearchText('login.php', AttributeValue) &&
       !kpmWideSearchText('capitalone.com', Url) &&
       !kpmWideSearchText('oszone.net', Url)
      ) ||
      (kpmWideSearchText('validate', AttributeValue) &&
       !kpmWideSearchText('flickr.com', Url)
      ) ||
      (kpmWideSearchText('submitAction', AttributeValue) &&
       kpmWideSearchText('login', AttributeValue)
      ) ||
      // real event handlers
      kpmWideSearchText('_login(', AttributeValue) ||
      kpmWideSearchText('setAction(', AttributeValue) ||
      kpmWideSearchText('submit(', AttributeValue) ||
      kpmWideSearchText('submitlogin(', AttributeValue) ||
      kpmWideSearchText('clicklogin(', AttributeValue) ||
      kpmWideSearchText('signin(', AttributeValue) ||
      kpmWideSearchText('authenticate(', AttributeValue) ||
      kpmWideSearchText('process_order_confirm(', AttributeValue) ||    // thinkgeek.com
      kpmWideSearchText("submitWithSwfEventId('login", AttributeValue) ||
      kpmWideSearchText('loginAccount(', AttributeValue)
     )
  {
	Result = true;
  }
  if (Result &&
	  (kpmWideSearchText('forgot', AttributeValue) ||
	   kpmWideSearchText('forget', AttributeValue)
	  )
	 )
  {
    Result = false;
  }
  // we must deny to press 'login with facebook' button
  if (Result && 
      kpmWideSearchText('facebook', AttributeValue)
	 )
  {
	Result = false;
  }
  return Result;
}

function kpmGetAdditionalFieldType(HtmlElement)
{
  var Result = kpm_afUnknown;
  if (HtmlElement)
  try
  {
    var ucTagName = '';
    var ucType = '';
    if (HtmlElement.tagName)
      ucTagName = kpmWidePrepareUpperCaseStr(HtmlElement.tagName);
    if (HtmlElement.type)
      ucType = kpmWidePrepareUpperCaseStr(HtmlElement.type);
    // FormElement
    if (ucTagName == kpm_szFORM)
    {
      Result = kpm_afHtmlForm;
    } 
    // ButtonElement or InputElement
    else if (ucTagName == kpm_szBUTTON ||
             ucTagName == kpm_szINPUT
            )
    {
      if (ucType == kpm_szBUTTON)
        Result = kpm_afButton; 
      else if (ucType == kpm_szIMAGE)
        Result = kpm_afButtonImage; 
      else if (ucType == kpm_szSUBMIT)
        Result = kpm_afButtonSubmit; 
      else if (ucType == kpm_szPASSWORD)
        Result = kpm_afEditPassword; 
      else if (ucType == kpm_szCHECKBOX)
        Result = kpm_afCheckBox; 
      else if (ucType == kpm_szRADIO)
        Result = kpm_afRadioButton;
     // support HTML 5 input types 
      else if ((ucType == kpm_szTEXT) || 
	           (ucType == 'EMAIL') || 
			   (ucType == 'TEL') || 
			   (ucType == 'MONTH') || 
			   (ucType == 'NUMBER') || 
			   (ucType == 'TIME') || 
			   (ucType == 'URL') || 
			   (ucType == 'WEEK') || 
			   (ucType == 'DATE'))
      {
        Result = kpm_afEdit;
        if (!kpmRealAdditionalFieldType &&
            kpmFocusingEditPasswordManager.IsFocusingEditPassword(HtmlElement)
           )
        {
          Result = kpm_afEditPassword;
        }
      }
    } 
    // TextAreaElement
    else if (ucTagName == kpm_szTEXTAREA)
    {
      Result = kpm_afMemo; 
    } 
    // SelectElement
    else if (ucTagName == kpm_szSELECT)
    {
      if ((ucType == kpm_szSELECT_MULTIPLE) ||
          (HtmlElement.size > 1))
      {
        Result = kpm_afListBox;
      } 
      else if ((ucType == kpm_szSELECT_ONE) ||
          (ucType == ''))
      {
        Result = kpm_afComboBoxDropDownList;
      }
    }
    // AnchorElement / ImgElement
    else if (ucTagName == kpm_szA || ucTagName == kpm_szIMG)
    {
      // check ID/Name
      if (kpmIsButtonSubmitIDorName(HtmlElement))
      {
        return kpm_afButtonSubmit;
      }
      var Url = kpmGetDocumentUrl(HtmlElement.ownerDocument);
      // check attributes
      for (var i = 0; i < 3; i++)
      {
        if (i > 0 && ucTagName != kpm_szA)
          break;
        var attributeName = '';
        var attributeValue = null;
        if (i == 0)
          attributeName = 'onclick'; 
        else if (i == 1)
        {
          if (ucTagName == kpm_szA)
            attributeName = 'href';
        }
        else if (i == 2)
        {
          if (ucTagName == kpm_szA)
            attributeValue = HtmlElement.innerHTML;
        }
        if (attributeName != '')
        {
          var funAttribute = HtmlElement.getAttribute(attributeName);
          if (funAttribute)
            attributeValue = funAttribute.toString();
        }
        if (kpmWideIsEmpty(attributeValue))
          continue;
        if (kpmIsButtonSubmitAttributeValue(attributeValue, Url))
        {
          return kpm_afButtonSubmit;
        }
      }
    }
  }
  catch (ErrorMessage)
  {
    Result = kpm_afUnknown;
    kpmLog.logError('kpmGetAdditionalFieldType() Error: ' + ErrorMessage);
  }
  return Result;
}

function kpmGetRealAdditionalFieldType(HtmlElement)
{
  kpmRealAdditionalFieldType = true;
  var Result = kpmGetAdditionalFieldType(HtmlElement);
  kpmRealAdditionalFieldType = false;
  return Result;
}

function kpmGetAdditionalFieldName(HtmlElement)
{
  var Result = kpm_afUnknown;
  if (HtmlElement)
  {
    Result = kpmGetElementName(HtmlElement);
    // if Name is empty then try to get it's ID
    if (kpmWideIsEmpty(Result))
    {
      Result = HtmlElement.id;
      // Opera bugfix - some forms returns their ids as an input element.
      if (Result && typeof Result != 'string')
        Result = HtmlElement.getAttribute('id');
    }
    // if ID is empty then create the unique name
    if (kpmWideIsEmpty(Result))
      Result = kpmWidePrepareLowerCaseStr(HtmlElement.tagName) + ' ' + kpmGetHTMLElementSourceIndex(HtmlElement);
  }
  return Result;
}

function kpmGetAdditionalFieldValue(HtmlElement)
{
  var Result = kpm_afUnknown;
  if (HtmlElement)
  {
    var ucTagName = '';
    if (HtmlElement.tagName)
      ucTagName = kpmWidePrepareUpperCaseStr(HtmlElement.tagName);
    // FormElement
    if (ucTagName == kpm_szFORM)
    {
      Result = HtmlElement.action;
      if (kpmWideIsEmpty(Result) ||
          kpmWideSameText(Result, '#') || 
          kpmWideSearchText('javascript:', Result) > 0
         )
        Result = HtmlElement.ownerDocument.URL;
    } else
    // ButtonElement
    if (ucTagName == kpm_szBUTTON)
    {
      Result = HtmlElement.innerHTML;
      // canonicalize FieldValue
      //   remove #0A chars
      if (typeof Result == 'string')
      {
        var s10 = String.fromCharCode(10);
        while (Result.indexOf(s10) >= 0)
          Result = Result.replace(s10, '');
      }
      //   trim spaces
      Result = kpmTrim(Result);
    } else
    // InputElement
    if (ucTagName == kpm_szINPUT)
    {
      var FieldType = kpmGetAdditionalFieldType(HtmlElement);
      if (FieldType == kpm_afButtonImage)
      {
        // for ButtonImage we must return src of <input type="image" src="...">
        Result = HtmlElement.src;
      } else
      if (FieldType == kpm_afCheckBox)
      {
        Result = HtmlElement.checked;
      } else
      { 
        Result = HtmlElement.value;
      }
    } else
    /*
    // AnchorElement
    if (ucTagName == kpm_szA)
    {
      Result = HtmlElement.innerHTML;
    } else
    // ImgElement
    if (ucTagName == kpm_szIMG)
    {
      Result = HtmlElement.alt;
    } else
    */
    // TextAreaElement or SelectElement
    {
      Result = HtmlElement.value;
    }
  }
  return Result;
}

function spHighlightHtmlElement(HtmlElement)
{
  if (!HtmlElement)
    return;
  HtmlElement.style.borderStyle = 'solid';
  HtmlElement.style.borderWidth = '1px';
  HtmlElement.style.borderTopColor = 'rgb(238, 50, 50)';
  HtmlElement.style.borderBottomColor = 'rgb(240, 80, 80)';
  HtmlElement.style.borderLeftColor = 'rgb(240, 80, 80)';
  HtmlElement.style.borderRightColor = 'rgb(240, 80, 80)';
  HtmlElement.style.boxShadow = '1px 1px rgb(250, 230, 230) inset'; // Default shaddow
  HtmlElement.style.MozBoxShadow = '1px 1px rgb(250, 230, 230) inset'; // Mozilla shaddow
  HtmlElement.style.WebkitBoxShadow = '1px 1px rgb(250, 230, 230) inset'; // Webkit shaddow
}

function kpmSetAdditionalFieldValue(HtmlElement, FieldValue)
{
  var Result = false;
  var FieldType = kpmGetAdditionalFieldType(HtmlElement);
  var FieldValueAsBoolean = FieldValue != '0';
  FieldValue = kpmWidePrepareStr(FieldValue);
  // InputElement
  if (kpmIsInputAdditionalFieldTypes(FieldType))
  {
    if (FieldType == kpm_afEdit)
    {
      // we must fire some events to allow fix Focusing Edit problem
      kpmHTMLElementFireEvent(HtmlElement, 'focus');
	  // in imo.im/ we have script, which clears login field after first 'change' event
      kpmHTMLElementFireEvent(HtmlElement, 'change');
      HtmlElement.value = FieldValue;
      kpmHTMLElementFireEvent(HtmlElement, 'change');
      kpmHTMLElementFireKeyEvent(HtmlElement, 'keypress', 0);
      kpmHTMLElementFireEvent(HtmlElement, 'blur');
      kpmHTMLElementFireEvent(HtmlElement, 'keyup');
      Result = kpmWideSameText(HtmlElement.value, FieldValue);
    }
    else if (FieldType == kpm_afEditPassword)
    {
      kpmHTMLElementFireEvent(HtmlElement, 'focus');
      HtmlElement.value = FieldValue;
      kpmHTMLElementFireEvent(HtmlElement, 'change');
      kpmHTMLElementFireKeyEvent(HtmlElement, 'keypress', 0);
      kpmHTMLElementFireEvent(HtmlElement, 'blur');
      kpmHTMLElementFireEvent(HtmlElement, 'keyup');
      Result = HtmlElement.value == FieldValue;
    }
    else if (FieldType == kpm_afCheckBox)
    {
      HtmlElement.checked = FieldValueAsBoolean;
      Result = HtmlElement.checked == FieldValueAsBoolean;
    }
    else if (FieldType == kpm_afRadioButton)
    {
      if (kpmWideSameText(HtmlElement.value, FieldValue))
      {
        HtmlElement.checked = true;
        Result = HtmlElement.checked == true;
      }
    }
  }
  // TextAreaElement
  else if (kpmIsTextAreaAdditionalFieldTypes(FieldType))
  {
    HtmlElement.value = FieldValue;
    Result = kpmWideSameText(HtmlElement.value, FieldValue);
  }
  // SelectElement
  else if (kpmIsSelectAdditionalFieldTypes(FieldType))
  {
    // check if
    try
    {
      var lcFieldValueAsText = kpmWidePrepareLowerCaseStr(FieldValue);
      if (lcFieldValueAsText.indexOf(kpm_FieldValueIsTextPrefix) == 0)
      {
        // Element's options
        for(var i = 0; i < HtmlElement.length; i++)
        {
          var Option = HtmlElement.options[i];
          var Text = kpmWidePrepareLowerCaseStr(Option.text);
          if (lcFieldValueAsText.indexOf(Text) == 1)
          {
            HtmlElement.selectedIndex = i;
            Result = true;
            break;
          }
        }
        return Result;
      }
    }
    catch (ErrorMessage)
    {
      kpmLog.logError('kpmSetAdditionalFieldValue(SelectElement) Error: ' + ErrorMessage);
    }
    // fill value
    HtmlElement.value = FieldValue;
    Result = kpmWideSameText(HtmlElement.value, FieldValue);
  }
  return Result;
}
  
function kpmIsAllowedAdditionalFieldForm(FormElement)
{
  // ignore empty & search forms
  if (FormElement &&
      FormElement.elements &&
      FormElement.elements.length &&
      kpmIsAllowedAdditionalFieldElement(FormElement, kpm_afHtmlForm, kpm_afAll)
     )
    return true;
  else
    return false;
}

function kpmIsAllowedAdditionalFieldElement(HtmlElement, FieldType, AllowHiddenFieldsType)
{
  var Result = false;
  if (HtmlElement &&
      (FieldType != kpm_afUnknown)
     )
  {
    if ((AllowHiddenFieldsType == kpm_afAll) ||
        kpmIsHTMLElementVisible(HtmlElement)
       )
    {
      if (kpmDebug_IsAllowedAdditionalFieldElement)
        kpmLog.logMessage('kpmIsAllowedAdditionalFieldElement() <' + 
          kpmGetElementName(HtmlElement) + '/' + HtmlElement.id + '> Checking additional allowing.');
      // we must ignore Search form & edit elements
      if (FieldType == kpm_afHtmlForm)
      {
        if (!kpmIsSearchAdditionalFieldElement(HtmlElement))
          Result = true;
        else if (kpmDebug_IsAllowedAdditionalFieldElement)
          kpmLog.logMessage('kpmIsAllowedAdditionalFieldElement() <' + 
            kpmGetElementName(HtmlElement) + '/' + HtmlElement.id + '> Search form!');  
        // INFO: we must deny to use additional forms, such as 
        // 'Find our friends with Skype' etc. (they all have 'ci_' prefix)
        // on facebook.com          
        var ADocument = HtmlElement.ownerDocument;          
        if (kpmCompareAdditionalFieldIDorName(HtmlElement, 'ci_') &&
          ADocument && kpmWideSearchText('facebook.com', kpmGetDocumentUrl(ADocument)))  
        Result = false;
        // Completely ignore hidden forms on twitter.com
        if (!kpmIsHTMLElementVisible(HtmlElement) &&
            kpmWideSearchText('twitter.com', kpmGetDocumentUrl(ADocument)))
          Result = false;
      } else
      if (FieldType == kpm_afEdit)
      {
        if (!kpmIsSearchAdditionalFieldElement(HtmlElement))
          Result = true; 
        else if (kpmDebug_IsAllowedAdditionalFieldElement)
          kpmLog.logMessage('kpmIsAllowedAdditionalFieldElement() <' + 
            kpmGetElementName(HtmlElement) + '/' + HtmlElement.id + '> Search element!');
      } else
      {
        Result = true;
      }
      // ignore readonly text
      if (FieldType == kpm_afEdit ||
          // INFO: Allow readonly edit password fields cause some sites are using virtual keyboard to enter password.
          // FieldType == kpm_afEditPassword ||
          FieldType == kpm_afMemo
         )
      {
        var readOnly = kpmIsHTMLElementReadOnly(HtmlElement);
        // INFO: Force allow readonly edit password on citibank.ru
        if (readOnly && FieldType == kpm_afEditPassword)
        {
           var ADocument = HtmlElement.ownerDocument;
           if (ADocument && kpmWideSearchText('citibank.ru/', kpmGetDocumentUrl(ADocument)))
             readOnly = false;
        }
        Result = Result && !readOnly;
        if (kpmDebug_IsAllowedAdditionalFieldElement && readOnly)
          kpmLog.logMessage('kpmIsAllowedAdditionalFieldElement() <' + 
            kpmGetElementName(HtmlElement) + '/' + HtmlElement.id + '> ReadOnly element.');
      }
      // ignore disabled text & password
      if (FieldType == kpm_afEdit ||
          FieldType == kpm_afEditPassword
         )
      {
        // disabled?
        var disabled = kpmIsHTMLElementDisabled(HtmlElement);
        Result = Result && !disabled;
        if (kpmDebug_IsAllowedAdditionalFieldElement && disabled)
          kpmLog.logMessage('kpmIsAllowedAdditionalFieldElement() <' + 
            kpmGetElementName(HtmlElement) + '/' + HtmlElement.id + '> Disabled element.');
      }    
      // ignore disabled & margined buttons
      if (FieldType == kpm_afButton ||
          FieldType == kpm_afButtonImage ||
          FieldType == kpm_afButtonSubmit
         )
      {
        // disabled?
        var disabled = kpmIsHTMLElementDisabled(HtmlElement);
        Result = Result && !disabled;
        if (kpmDebug_IsAllowedAdditionalFieldElement && disabled)
          kpmLog.logMessage('kpmIsAllowedAdditionalFieldElement() <' + 
            kpmGetElementName(HtmlElement) + '/' + HtmlElement.id + '> Disabled element.');
        // is search?
        var isSearch = kpmIsSearchAdditionalFieldElement(HtmlElement);
        Result = Result && !isSearch;
        if (kpmDebug_IsAllowedAdditionalFieldElement && isSearch)
          kpmLog.logMessage('kpmIsAllowedAdditionalFieldElement() <' + 
            kpmGetElementName(HtmlElement) + '/' + HtmlElement.id + '> Search element.');
        // margined?
        var margined = false;
        var ADocument = HtmlElement.ownerDocument;
        if (ADocument == null)
          ADocument = document;
        if (ADocument == null)
          return true;
        //   check currentStyle
        if (ADocument.defaultView && ADocument.defaultView.getComputedStyle)
        try
        {
          var computedStyle = ADocument.defaultView.getComputedStyle(HtmlElement, '');
          if (computedStyle)
          {
            // check currentStyle.display value
            var marginLeft = parseInt(computedStyle.marginLeft);
            if (marginLeft < -100)
              margined = true;
            if (kpmDebug_IsAllowedAdditionalFieldElement && margined)
              kpmLog.logMessage('kpmIsAllowedAdditionalFieldElement() <' + 
                kpmGetElementName(HtmlElement) + '/' + HtmlElement.id + '> Margined-left element: <' + computedStyle.marginLeft + '>'
              );
          }
        }
        catch (ErrorMessage)
        {
          kpmLog.logError('kpmIsAllowedAdditionalFieldElement() GetComputedStyle() error: ' + ErrorMessage);
        }
        Result = Result && !margined;
      }
    }
    else
    {
      if (kpmDebug_IsAllowedAdditionalFieldElement)
        kpmLog.logMessage('kpmIsAllowedAdditionalFieldElement() <' + 
          kpmGetElementName(HtmlElement) + '/' + HtmlElement.id + '> Hidden element.');
    }
  }
  if (kpmDebug_IsAllowedAdditionalFieldElement && !Result)
    kpmLog.logMessage('kpmIsAllowedAdditionalFieldElement() <' + 
      kpmGetElementName(HtmlElement) + '/' + HtmlElement.id + '> Disallowed element.');
  return Result;  
}

function kpmIsSearchAdditionalFieldElement(HtmlElement)
{
  // search	
  if ((kpmCompareAdditionalFieldIDorName(HtmlElement, 'search') ||
	   kpmCompareAdditionalFieldIDorName(HtmlElement, 'find')) &&
      // on officialpayments.com exists login field with 'userName' name & 'search' id!
      (!kpmCompareAdditionalFieldIDorName(HtmlElement, 'userName'))
     )
  {
    return true;
  }
  // txtIssue
  if (kpmCompareAdditionalFieldIDorName(HtmlElement, 'txtIssue'))
  {
    return true;
  }
  // CtrlVerticalNav1_txtProduct on northwestpharmacy.com
  if (kpmCompareAdditionalFieldIDorName(HtmlElement, 'CtrlVerticalNav1_txtProduct'))
  {
    return true;
  }
  // HTML 5 search element type 
  if (kpmGetAdditionalFieldType(HtmlElement) == kpm_afEdit && 
      kpmWidePrepareUpperCaseStr(HtmlElement.type) == 'SEARCH') 
  {
	return true;
  }
  return false;
}

function kpmCompareAdditionalFieldIDorName(HtmlElement, AValue)
{
  if (HtmlElement && !kpmWideIsEmpty(AValue))
  {
    if (kpmWideSearchText(AValue, kpmGetElementName(HtmlElement)) ||
        kpmWideSearchText(AValue, HtmlElement.id)
       )
    {
      return true;
    }
  }
  return false;
}

function kpmCompareSameAdditionalFieldIDorName(HtmlElement, AValue)
{
  if (HtmlElement && !kpmWideIsEmpty(AValue))
  {
    if (kpmWideSameText(AValue, kpmGetElementName(HtmlElement)) ||
        kpmWideSameText(AValue, HtmlElement.id)
       )
    {
      return true;
    }
  }
  return false;
}

function kpmIsTimeredAutofillUrl(Url)
{
  if (kpmWideSearchText('schwab.com', Url) ||
      kpmWideSearchText('login.live.com', Url) ||
      kpmWideSearchText('turbotaxonline.intuit.com', Url) ||
      kpmWideSearchText('pip.verisignlabs.com', Url) ||
      kpmWideSearchText('personal.vanguard.com', Url) ||
      kpmWideSearchText('o2online.de', Url) ||
      kpmWideSearchText('mjiteamworks.com', Url) ||
      kpmWideSearchText('accountonline.com', Url) ||
      kpmWideSearchText('lamantine.unfuddle.com', Url) ||
      kpmWideSearchText('onstar.com', Url) ||
      kpmWideSearchText('betfair.com', Url) ||
      kpmWideSearchText('hsbc.co.uk', Url) ||
      kpmWideSearchText('alfabank.ru', Url) || 
      kpmWideSearchText('bmoinvestorline.com', Url) || 
	    kpmWideSearchText('imo.im', Url) ||  
	    kpmWideSearchText('quakelive.com', Url) || 
	    kpmWideSearchText('walletone.com', Url) ||  
	    kpmWideSearchText('pandora.com', Url) ||   
 	    kpmWideSearchText('my.vudu.com', Url) || 
	    kpmWideSearchText('alibaba.com', Url) || 
	    kpmWideSearchText('equabanking.cz', Url) || 
	  kpmWideSearchText('interactivebrokers.ch', Url) ||           	    
      kpmWideSearchText('memeo.com', Url) 
     )
    return true;
  else
    return false;
}

function kpmGetTimeredAutofillUrlInterval(Url)
{
  var Result = 500;  // ~ 0.5 sec
  if (kpmWideSearchText('alfabank.ru', Url) ||
      kpmWideSearchText('pandora.com', Url) ||
      kpmWideSearchText('equabanking.cz', Url) || 
	  kpmWideSearchText('interactivebrokers.ch', Url) ||  	  
      kpmWideSearchText('quakelive.com', Url))
  {
    // we need to use this interval to allow the document dynamically create it's contents
    Result = 2000; // ~ 2 sec
  }
  return Result;
}

// TkpmBeforeSubmitElement ------------------------------------------------------

function TkpmBeforeSubmitElement()
{
  this.NodeName = '';
  this.HtmlFormName = '';
  this.HtmlFormAction = '';
  this.Name = '';
  this.OldValue = '';
  this.NewValue = '';
}
    
TkpmBeforeSubmitElement.prototype =
{
  IsEmpty: function()
  {
    var Result =
      kpmWideIsEmpty(this.HtmlFormName) &&
      kpmWideIsEmpty(this.HtmlFormAction) &&
      kpmWideIsEmpty(this.Name) &&
      // ?? kpmWideIsEmpty(this.OldValue) &&
      kpmWideIsEmpty(this.NewValue);
    if (this.NodeName == 'login')
    {
      // LoginElement
      Result =
        Result ||
        kpmWideSameText(this.NewValue, this.OldValue) ||
        kpmWideIsEmpty(this.NewValue);
    } else
    {
      // PasswordElement
      Result =
        Result ||
        (this.NewValue == this.OldValue) ||
        (this.NewValue == '');
    }
    return Result;
  },

  Clear: function()
  {
    this.HtmlFormName = '';
    this.HtmlFormAction = '';
    this.Name = '';
    this.OldValue = '';
    this.NewValue = '';   
  },
  
  LoadFromXMLDocument: function(XMLDocument)
  {
    if (XMLDocument == null)
      return false;
    // get <XMLElement>
    var XMLElement = XMLDocument.getElementsByTagName(this.NodeName);
    if (XMLElement.length == 1)
      XMLElement = XMLElement[0];
    else
      return false;
    // load from XMLElement
    this.HtmlFormName = XMLElement.getAttribute('HtmlFormName');
    if (this.HtmlFormName == null)
      this.HtmlFormName = '';
    this.HtmlFormAction = XMLElement.getAttribute('HtmlFormAction');
    if (this.HtmlFormAction == null)
      this.HtmlFormAction = '';
    this.Name = XMLElement.getAttribute('Name');
    if (this.Name == null)
      this.Name = '';
    this.OldValue = XMLElement.getAttribute('OldValue');
    if (this.OldValue == null)
      this.OldValue = '';
    this.NewValue = XMLElement.getAttribute('NewValue');
    if (this.NewValue == null)
      this.NewValue = '';
    return true;
  }
}

// TkpmBeforeSubmitManager ------------------------------------------------------

function TkpmBeforeSubmitManager()
{
  this.LoginElement = new TkpmBeforeSubmitElement();
  this.PasswordElement = new TkpmBeforeSubmitElement();
  this.LoginElement.NodeName = 'login';
  this.PasswordElement.NodeName = 'password';
}

TkpmBeforeSubmitManager.prototype =
{
  IsEmpty: function()
  {
    return this.LoginElement.IsEmpty() && this.PasswordElement.IsEmpty();
  },

  LoadFromXML: function(XML)
  {
    this.LoginElement.Clear();
    this.PasswordElement.Clear();
    
    // check if XML is empty string
    if (kpmWideIsEmpty(XML))
      return false;
      
    // create new XMLDocument & load it from external XML
    try
    {
      var XMLParser = kpmNewXMLParser();
      var XMLDocument = XMLParser.parseFromString(XML, "text/xml");
      this.LoginElement.LoadFromXMLDocument(XMLDocument);
      this.PasswordElement.LoadFromXMLDocument(XMLDocument);
    }
    catch (ErrorMessage)
    {
      kpmLog.logError('TkpmBeforeSubmitManager.LoadFromXML() Error: ' + ErrorMessage);
      return false;
    }
    return true;
  }
}

// BeforeSubmit Autofill -------------------------------------------------------

function kpmBeforeSubmitFillElement(ADocument, AForm, AElement, ABeforeSubmitElement)
{
  var Result =
    (!ABeforeSubmitElement.IsEmpty()) &&
    kpmWideSameText(kpmGetAdditionalFieldName(AForm), ABeforeSubmitElement.HtmlFormName) &&
    kpmWideSameText(kpmGetAdditionalFieldValue(AForm), ABeforeSubmitElement.HtmlFormAction) &&
    kpmWideSameText(kpmGetAdditionalFieldName(AElement), ABeforeSubmitElement.Name);
  if (Result)
  {
    if (kpmSetAdditionalFieldValue(AElement, ABeforeSubmitElement.NewValue))
      ABeforeSubmitElement.Clear();
  }  
}

function kpmFillBeforeSubmitWindow(AWindow, ABeforeSubmitManager)
{
  if (AWindow && ABeforeSubmitManager)
  try
  {
    var DocumentList = kpmGetHTMLDocumentList(AWindow);
    for (var i = 0; i < DocumentList.length; i++)
    {
      var Document = DocumentList[i];
      var DocumentFilled = false;
      // try to autofill the Document's form
      if (Document.forms)
      {
        for(var j = 0; j < Document.forms.length; j++)
        {
          var FormElement = Document.forms[j];
          // check if form is allowed?
          if (!kpmIsAllowedAdditionalFieldForm(FormElement))
            continue;
          // try to search form's elements
          for(var k = 0; k < FormElement.elements.length; k++)
          {
            var Element = FormElement.elements[k];
            // get element's additional field type 
            var FieldType = kpmGetAdditionalFieldType(Element);
            // ignore disallowed additional field elements
            if (!kpmIsAllowedAdditionalFieldElement(Element, FieldType, kpm_afAll))
              continue;
            if (FieldType == kpm_afEdit || FieldType == kpm_afEditPassword)
            {
              if (kpmBeforeSubmitFillElement(Document, FormElement, Element, ABeforeSubmitManager.LoginElement))
                DocumentFilled = true;
              if (kpmBeforeSubmitFillElement(Document, FormElement, Element, ABeforeSubmitManager.PasswordElement))
                DocumentFilled = true;
              if (ABeforeSubmitManager.IsEmpty())
                return true;
            }
          }
        }
      }
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmFillBeforeSubmitWindow() Error: ' + ErrorMessage);
  }
}

// TkpmMultipageAuthorizationLoginInfo ------------------------------------------

function TkpmMultipageAuthorizationLoginInfo()
{
  this.GeneralDomainName = '';
  this.FormIDorName = '';
  this.ElementIDorName = '';
  this.HiddenElementIDorName = '';
}

TkpmMultipageAuthorizationLoginInfo.prototype =
{
  // WARNING: We mustn't declare here variables, cause when pass this object
  //   into function - variables would be reset to default values!
}

// TkpmMultipageAuthorizationManager --------------------------------------------

function TkpmMultipageAuthorizationManager()
{
  this.Items = new Array();
  var Item = null;
  // bankofamerica.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'bankofamerica.com';
  Item.FormIDorName = 'frmSignIn';
  Item.ElementIDorName = 'id';
  Item.HiddenElementIDorName = 'Access_ID';
  // netxselect.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'netxselect.com';
  Item.FormIDorName = 'Login';
  Item.ElementIDorName = 'CustId';
  Item.HiddenElementIDorName = '';
  // rbsdigital.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'rbsdigital.com';
  Item.FormIDorName = 'form';
  Item.ElementIDorName = 'DBID_edit';
  Item.HiddenElementIDorName = '';
  // tel3advantage.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'tel3advantage.com';
  Item.FormIDorName = 'form1';
  Item.ElementIDorName = 'username';
  Item.HiddenElementIDorName = '';
  // unionbank.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'unionbank.com';
  Item.FormIDorName = 'homepage';
  Item.ElementIDorName = 'userId';
  Item.HiddenElementIDorName = '';
  // uboc.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'uboc.com';
  Item.FormIDorName = 'homepage';
  Item.ElementIDorName = 'userId';
  Item.HiddenElementIDorName = '';
  // zionsbank.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'zionsbank.com';
  Item.FormIDorName = 'logonForm';
  Item.ElementIDorName = 'publicCred1';
  Item.HiddenElementIDorName = '';
  // nwolb.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'nwolb.com';
  Item.FormIDorName = 'aspnetForm';
  Item.ElementIDorName = 'ctl00_mainContent_LI5TABA_DBID_edit';
  Item.HiddenElementIDorName = '';
  // bmo.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'bmo.com';
  Item.FormIDorName = 'SignOn';
  Item.ElementIDorName = 'FBC_Number';
  Item.HiddenElementIDorName = '';
  // hsbc.co.uk
  Item = this.AddItem();
  Item.GeneralDomainName = 'hsbc.co.uk';
  Item.FormIDorName = 'IBloginForm';
  Item.ElementIDorName = 'intbankingID';
  Item.HiddenElementIDorName = '';
}

TkpmMultipageAuthorizationManager.prototype =
{
  AddItem: function()
  {
    var Result = new TkpmMultipageAuthorizationLoginInfo();
    this.Items[this.Items.length] = Result;
    return Result;
  },

  IsMultipageAuthorizationDocument: function(ADocument)
  {
    var URL = kpmGetDocumentUrl(ADocument);
    for (var i = 0; i < this.Items.length; i++)
    {
      var Info = this.Items[i];
      if (kpmWideSearchText(Info.GeneralDomainName, URL))
        return true;
    }
    return false;
  },

  IsMultipageAuthorizationLogin: function(FormElement, ElementList, Element, MustBeFilled, IsLoginElement, Login)
  {
    try
    {
      IsLoginElement.value = false;
      if (FormElement && ElementList && Element && kpmWideIsEmpty(Login))
      {
        // 1. use default id/name-search engine
        var FormAction = kpmGetAdditionalFieldValue(FormElement);
        for (var i = 0; i < this.Items.length; i++)
        {
          var Info = this.Items[i];
          if (kpmWideSearchText(Info.GeneralDomainName, FormAction) &&
              kpmCompareAdditionalFieldIDorName(FormElement, Info.FormIDorName)
             )
          {
            if (!kpmWideIsEmpty(Info.HiddenElementIDorName))
            {
              // HiddenElement
              if (kpmCompareAdditionalFieldIDorName(Element, Info.HiddenElementIDorName))
              {
                Login.value = kpmGetAdditionalFieldValue(Element);
                if (!kpmWideIsEmpty(Login))
                {
                  if (kpmDebug_IsMultipageAuthorizationLogin)
                    kpmLog.logMessage('TkpmMultipageAuthorizationManager.IsMultipageAuthorizationLogin() Login: <' + Login.value + '>');
                  return true;
                }
              }
            }
            // Element
            if (kpmCompareAdditionalFieldIDorName(Element, Info.ElementIDorName))
            {
              IsLoginElement.value = true;
              Login.value = kpmGetAdditionalFieldValue(Element);
              if (!kpmWideIsEmpty(Login))
              {
                if (kpmDebug_IsMultipageAuthorizationLogin)
                  kpmLog.logMessage('TkpmMultipageAuthorizationManager.IsMultipageAuthorizationLogin() Login: <' + Login.value + '>');
                return true;
              }
            }
          }
        }
        // check only for afEdit and not empty element!
        if (kpmGetAdditionalFieldType(Element) != kpm_afEdit)
          return false;
        var tmpLogin = kpmGetAdditionalFieldValue(Element);
        if (MustBeFilled && kpmWideIsEmpty(tmpLogin))
          return false;
        if (!kpmAutofill())
          return false;
        // 2. use autofind engine
        var EditCount = 0;
        for (var i = 0; i < ElementList.length; i++)
        {
          var FieldType = kpmGetAdditionalFieldType(ElementList[i]);
          if (FieldType == kpm_afEdit)
            EditCount++; else
          if (FieldType == kpm_afEditPassword)
            return false;
        }
        if (EditCount == 1)
        {
          var Document = Element.ownerDocument;
          if (kpmAutofill().IsMultipageAuthorizationLogin(Document.URL, 
            kpmGetAdditionalFieldValue(FormElement), Element.id, kpmGetElementName(Element)))
          {
            IsLoginElement.value = true;
            Login.value = tmpLogin;
            if (!kpmWideIsEmpty(Login))
            {
              if (kpmDebug_IsMultipageAuthorizationLogin)
                kpmLog.logMessage('TkpmMultipageAuthorizationManager.IsMultipageAuthorizationLogin() Login: <' + Login.value + '>');
              return true;
            }
          }
        }
      }
      return false;
    }
    catch (ErrorMessage)
    {
      kpmLog.logError('TkpmMultipageAuthorizationManager.IsMultipageAuthorizationLogin() Error: ' + ErrorMessage);
    }
  }
}

var kpmMultipageAuthorizationManager = new TkpmMultipageAuthorizationManager();

// TkpmFocusingElementsInfo -----------------------------------------------------

function TkpmFocusingElementsInfo()
{
  this.GeneralDomainName = '';
  this.LoginElementIDorName = '';
  this.PasswordElementIDorName = '';
}

TkpmFocusingElementsInfo.prototype =
{
  // WARNING: We mustn't declare here variables, cause when pass this object
  //   into function - variables would be reset to default values!
}

// TkpmFocusingEditPasswordManager ----------------------------------------------

function TkpmFocusingEditPasswordManager()
{
  this.Items = new Array();
  var Item = null;
  // juventus-team.hu
  Item = this.AddItem();
  Item.GeneralDomainName = 'juventus-team.hu';
  Item.LoginElementIDorName = 'ctl00_UCO_login1_EDI_loginnev';
  Item.PasswordElementIDorName = 'ctl00_UCO_login1_EDI_jelszo';
  // facebook.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'facebook.com';
  Item.LoginElementIDorName = '';
  Item.PasswordElementIDorName = 'pass_placeholder';
  // alltel.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'alltel.com';
  Item.LoginElementIDorName = 'tmpUsername';
  Item.PasswordElementIDorName = 'tmpPwd';
  // vanguard.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'vanguard.com';
  Item.LoginElementIDorName = 'USER';
  Item.PasswordElementIDorName = '';
  // troweprice.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'troweprice.com';
  Item.LoginElementIDorName = 'usernameInput';
  Item.PasswordElementIDorName = '';
  // asda.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'asda.com';
  Item.LoginElementIDorName = 'h-signin-input1';
  Item.PasswordElementIDorName = 'h-signin-input3';
  // janus.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'janus.com';
  Item.LoginElementIDorName = 'txtssn';
  Item.PasswordElementIDorName = 'txtpin';
  // gamevance.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'gamevance.com';
  Item.LoginElementIDorName = 'mem_u';
  Item.PasswordElementIDorName = 'mem_p_temp';
  // aeroplan.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'aeroplan.com';
  Item.LoginElementIDorName = '';
  Item.PasswordElementIDorName = 'passwordField-watermark';
  // stumbleupon.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'stumbleupon.com';
  Item.LoginElementIDorName = 'username';
  Item.PasswordElementIDorName = 'password';
  // mybroadview.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'mybroadview.com';
  Item.LoginElementIDorName = 'ctl00_m_g_153fb1c5_54d5_40ba_9d13_113ee57f6769_txtUserName';
  Item.PasswordElementIDorName = 'ctl00_m_g_153fb1c5_54d5_40ba_9d13_113ee57f6769_txtPasswordDummy';
  // friendfinder.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'friendfinder.com';
  Item.LoginElementIDorName = '';
  Item.PasswordElementIDorName = 'pwx';
  // vodafone.de
  Item = this.AddItem();
  Item.GeneralDomainName = 'vodafone.de';
  Item.LoginElementIDorName = 'inetaccount';
  Item.PasswordElementIDorName = 'inetinfo';
  // softwareload.de / telekom.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'telekom.com';
  Item.LoginElementIDorName = 'usr';
  Item.PasswordElementIDorName = 'pwdText';
  /*
  // gmx.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'gmx.com';
  Item.LoginElementIDorName = 'TextfieldEmail';
  Item.PasswordElementIDorName = '_id2_';
  */
  // pinebrookgolfclub.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'pinebrookgolfclub.com';
  Item.LoginElementIDorName = 'memberId';
  Item.PasswordElementIDorName = 'pass_temp';
  // haustechnikdialog.de
  Item = this.AddItem();
  Item.GeneralDomainName = 'haustechnikdialog.de';
  Item.LoginElementIDorName = 'ctl01_cphHeadDiv_usrStatus_lvMain_Login1_UserName_text';
  Item.PasswordElementIDorName = 'ctl01_cphHeadDiv_usrStatus_lvMain_Login1_RadTextBox1_text';
  // win2day.at
  Item = this.AddItem();
  Item.GeneralDomainName = 'win2day.at';
  Item.LoginElementIDorName = 'usernameInput';
  Item.PasswordElementIDorName = 'defaultpwd';
  // betfair.com
  Item = this.AddItem();
  Item.GeneralDomainName = 'betfair.com';
  Item.LoginElementIDorName = 'username';
  Item.PasswordElementIDorName = 'passwordText';
  // navitel.su
  Item = this.AddItem();
  Item.GeneralDomainName = 'navitel.su';
  Item.LoginElementIDorName = '';
  Item.PasswordElementIDorName = 'password';
  // navitel.su
  Item = this.AddItem();
  Item.GeneralDomainName = 'imo.im';
  Item.LoginElementIDorName = 'splash_imo_username';
  Item.PasswordElementIDorName = 'splash_imo_password';
}

TkpmFocusingEditPasswordManager.prototype =
{
  AddItem: function()
  {
    var Result = new TkpmFocusingElementsInfo();
    this.Items[this.Items.length] = Result;
    return Result;
  },
  
  IsFocusingEdit: function (AElement, AIsLoginEdit)
  {
    if (AElement)
    try
    {
      var FormAction = kpmGetAdditionalFieldValue(AElement.form);
      for (var i = 0; i < this.Items.length; i++)
      {
        var Info = this.Items[i];
        if (kpmWideSearchText(Info.GeneralDomainName, FormAction))
        {
          var ElementIDorName = '';
          if (AIsLoginEdit)
            ElementIDorName = Info.LoginElementIDorName
          else
            ElementIDorName = Info.PasswordElementIDorName;
          if (kpmCompareSameAdditionalFieldIDorName(AElement, ElementIDorName))
            return true;
        }
      }
    }
    catch (ErrorMessage)
    {
      kpmLog.logError('TkpmFocusingEditPasswordManager.IsFocusingEdit() Error: ' + ErrorMessage);
    }
    return false;
  },

  IsFocusingEditPassword: function(AElement)
  {
    return this.IsFocusingEdit(AElement, false);
  },
  
  FixFocusingEdit: function (AElement)
  {
    if (AElement)
    try
    {
      // get element's real additional field type
      var FieldType = kpmGetRealAdditionalFieldType(AElement);
      // we can fix only visible afEdit element which can later switch to afEdit or afEditPassword!
      if ((FieldType == kpm_afEdit) &&
          kpmIsHTMLElementVisible(AElement) &&
          (this.IsFocusingEdit(AElement, false) || // password
           this.IsFocusingEdit(AElement, true)     // login
          )
         )
      {
        // log before info
        if (kpmDebug_FixFocusingEdit)
          kpmLog.logMessage('TkpmFocusingEditPasswordManager.FixFocusingEdit() Element <' + kpmGetElementName(AElement) + '/' + AElement.id + '>');
        // call events
        kpmHTMLElementFireEvent(AElement, 'blur');
        kpmHTMLElementFireEvent(AElement, 'focus');
        kpmHTMLElementFireEvent(AElement, 'change');
        kpmHTMLElementFireMouseEvent(AElement, 'click');
        kpmHTMLElementFireMouseEvent(AElement, 'mousedown');
        kpmHTMLElementFireKeyEvent(AElement, 'keydown', 0);
        // try to set focus on edit
        kpmHTMLElementSetFocus(AElement);
        // log after info
        if (kpmDebug_FixFocusingEdit)
          kpmLog.logMessage('TkpmFocusingEditPasswordManager.FixFocusingEdit() Element after fix visible: ' + kpmIsHTMLElementVisible(AElement));
        // finally element mustn't be visible
        return kpmIsHTMLElementVisible(AElement);
      }
    }
    catch (ErrorMessage)
    {
      kpmLog.logError('TkpmFocusingEditPasswordManager.FixFocusingEdit() Error: ' + ErrorMessage);
    }
    return false;
  },
  
  FixFocusingEditList: function (AElementList)
  {
    if (AElementList)
    try
    {
      // fix focusing edit login/password
      var MustSortList = false;
      for(var i = 0; i < AElementList.length; i++)
      {
        var Element = AElementList[i];
        if (this.FixFocusingEdit(Element))
          MustSortList = true;
      }
      // sort the list if need
      if (MustSortList)
        AElementList.sort(kpmHTMLElementListCompare);
    }
    catch (ErrorMessage)
    {
      kpmLog.logError('TkpmFocusingEditPasswordManager.FixFocusingEditList() Error: ' + ErrorMessage);
    }
  }
}

var kpmFocusingEditPasswordManager = new TkpmFocusingEditPasswordManager();

// TkpmAdditionalField ----------------------------------------------------------

function TkpmAdditionalField()
{
  // common
  this.FieldType = kpm_afUnknown;
  this.FieldName = '';
  this.FieldValue = '';
  this.AutofillValueType = kpm_atValue;
  // used for web-pages only
  this.HtmlDocumentUrl = '';
  this.HtmlFormName = '';
  this.HtmlFormAction = '';
  // runtime only
  this.Autofilled = 0;
}
    
TkpmAdditionalField.prototype =
{
  Initialized: function()
  {
    return this.FieldType != kpm_afUnknown;
  },

  FieldValueAsBoolean: function()
  {
    return this.FieldValue != '0';
  },

  LoadFromXMLElement: function(XMLElement)
  {
    if (XMLElement == null)
      return false;
    // common
    this.FieldType = XMLElement.getAttribute(kpm_szFieldType);
    if (this.FieldType == null)
      this.FieldType = kpm_afUnknown;
    this.FieldName = XMLElement.getAttribute(kpm_szFieldName);
    if (this.FieldName == null)
      this.FieldName = '';
    this.FieldValue = XMLElement.getAttribute(kpm_szFieldValue);
    if (this.FieldValue == null)
      this.FieldValue = '';
    this.AutofillValueType = XMLElement.getAttribute(kpm_szAutofillValueType);
    if (this.AutofillValueType == null)
      this.AutofillValueType = kpm_atValue;
    // used for web-pages only
    this.HtmlDocumentUrl = XMLElement.getAttribute(kpm_szHtmlDocumentUrl);
    if (this.HtmlDocumentUrl == null)
      this.HtmlDocumentUrl = '';
    this.HtmlFormName = XMLElement.getAttribute(kpm_szHtmlFormName);
    if (this.HtmlFormName == null)
      this.HtmlFormName = '';
    this.HtmlFormAction = XMLElement.getAttribute(kpm_szHtmlFormAction);
    if (this.HtmlFormAction == null)
      this.HtmlFormAction = '';
    return true;
  },

  InitializeFromHTMLElement: function(HTMLDocument, HTMLForm, HTMLElement)
  {
    // common
    this.FieldType = kpmGetAdditionalFieldType(HTMLElement);
    this.FieldName = kpmGetAdditionalFieldName(HTMLElement);
    this.FieldValue = kpmGetAdditionalFieldValue(HTMLElement);
    // used for web-pages only    
    this.HtmlDocumentUrl = HTMLDocument.location.href;
    if (HTMLForm)
    {
      this.HtmlFormName = kpmGetAdditionalFieldName(HTMLForm);
      this.HtmlFormAction = kpmGetAdditionalFieldValue(HTMLForm);
    } else
    {
      this.HtmlFormName = '';
      this.HtmlFormAction = kpmWidePrepareLowerCaseStr(kpmGetDocumentUrl(HTMLDocument));
    }
  },

  Compare: function(DestAdditionalField, opExcludeFieldValueCompare)
  {
    if (DestAdditionalField == null)
      return false;
    /*
    alert(    
      'this.AutofillValueType: ' + this.AutofillValueType + kpmLineBreak +
      'dest.AutofillValueType: ' + DestAdditionalField.AutofillValueType + kpmLineBreak +
      'this.FieldType: ' + this.FieldType + kpmLineBreak +
      'dest.FieldType: ' + DestAdditionalField.FieldType + kpmLineBreak +
      'this.FieldName: ' + this.FieldName + kpmLineBreak +
      'dest.FieldName: ' + DestAdditionalField.FieldName + kpmLineBreak +
      'this.FieldValue: ' + this.FieldValue + kpmLineBreak +
      'dest.FieldValue: ' + DestAdditionalField.FieldValue + kpmLineBreak +
      'this.HtmlDocumentUrl: ' + this.HtmlDocumentUrl + kpmLineBreak +
      'dest.HtmlDocumentUrl: ' + DestAdditionalField.HtmlDocumentUrl + kpmLineBreak +
      'this.HtmlFormName: ' + this.HtmlFormName + kpmLineBreak +
      'dest.HtmlFormName: ' + DestAdditionalField.HtmlFormName + kpmLineBreak +
      'this.HtmlFormAction: ' + this.HtmlFormAction + kpmLineBreak +
      'dest.HtmlFormAction: ' + DestAdditionalField.HtmlFormAction
    );
    // */
    var Result =
      // check FieldType
      (this.FieldType != kpm_afUnknown) &&
      (DestAdditionalField.FieldType != kpm_afUnknown) &&
      (this.FieldType == DestAdditionalField.FieldType) &&
      // check FieldName
      kpmWideSameText(this.FieldName, DestAdditionalField.FieldName) &&
      // if need - check FieldValue
      ((this.FieldType == kpm_afHtmlForm) ||
       opExcludeFieldValueCompare ||
       kpmWideIsEmpty(this.FieldValue) ||
       kpmWideSameText(this.FieldValue, DestAdditionalField.FieldValue)
      ) &&
      // check HtmlDocumentUrl or domain of HtmlDocumentUrl
      kpmCompareUrl(this.HtmlDocumentUrl, DestAdditionalField.HtmlDocumentUrl) &&
      // check HtmlFormName
      (kpmWideIsEmpty(this.HtmlFormName) ||
       // ?? kpmWideIsEmpty(DestAdditionalField.HtmlFormName) ||
       kpmWideSameText(this.HtmlFormName, DestAdditionalField.HtmlFormName)
      ) &&
      // check HtmlFormAction
      (kpmWideIsEmpty(this.HtmlFormAction) ||      
       // ?? kpmWideIsEmpty(DestAdditionalField.HtmlFormAction) ||
       kpmCompareUrl(this.HtmlFormAction, DestAdditionalField.HtmlFormAction)
      );
    return Result;
  },
  
  SetElementValue: function(Element)
  {
    this.Autofilled = kpmSetAdditionalFieldValue(Element, this.FieldValue);
    return this.Autofilled;
  }
}

// TkpmAdditionalFieldsManager --------------------------------------------------

function TkpmAdditionalFieldsManager()
{
  this.FormSubmitField = new TkpmAdditionalField();
  this.ButtonSubmitField = new TkpmAdditionalField();
  // runtime only
  this.ButtonCheckAvailabilityField = new TkpmAdditionalField();
  this.FocusingField = new TkpmAdditionalField();
  this.Items = null;
}

TkpmAdditionalFieldsManager.prototype =
{
  LoadFromXml: function(AXml)
  {
    // check if AXml is empty string
    if (kpmWideIsEmpty(AXml))
      return false;
    // create new XMLDocument & load it from external Xml
    try
    {
      if (!document.implementation || !document.implementation.createDocument)
        return false;
      var XMLParser = kpmNewXMLParser();
      var XMLDocument = XMLParser.parseFromString(AXml, "text/xml");
    }
    catch (ErrorMessage)
    {
      kpmLog.logError('TkpmAdditionalFieldsManager.LoadFromXml() Error: ' + ErrorMessage);
      return false;
    }
    // get <FormSubmitField>
    var xmlFormSubmitField = XMLDocument.getElementsByTagName(kpm_szFormSubmitField);
    if (xmlFormSubmitField.length == 1)
      xmlFormSubmitField = xmlFormSubmitField[0];
    else
      xmlFormSubmitField = null;
    // get <ButtonSubmitField>
    var xmlButtonSubmitField = XMLDocument.getElementsByTagName(kpm_szButtonSubmitField);
    if (xmlButtonSubmitField.length == 1)
      xmlButtonSubmitField = xmlButtonSubmitField[0];
    else
      xmlButtonSubmitField = null;
    // get <ButtonCheckAvailabilityField>
    var xmlButtonCheckAvailabilityField = XMLDocument.getElementsByTagName(kpm_szButtonCheckAvailabilityField);
    if (xmlButtonCheckAvailabilityField.length == 1)
      xmlButtonCheckAvailabilityField = xmlButtonCheckAvailabilityField[0];
    else
      xmlButtonCheckAvailabilityField = null;
    // get <FocusingField>
    var xmlFocusingField = XMLDocument.getElementsByTagName(kpm_szFocusingField);
    if (xmlFocusingField.length == 1)
      xmlFocusingField = xmlFocusingField[0];
    else
      xmlFocusingField = null;
    // get <AdditionalField> nodes
    var xmlAdditionalFields = XMLDocument.getElementsByTagName(kpm_szAdditionalField);
    // check for any AdditionalFields specified?
    if ((xmlFormSubmitField == null) &&
        (xmlButtonSubmitField == null) &&
        (xmlAdditionalFields.length  == 0)
       )
    {
      return false;
    }
    // load FormSubmitField, ButtonSubmitField & Items
    this.FormSubmitField.LoadFromXMLElement(xmlFormSubmitField);
    this.ButtonSubmitField.LoadFromXMLElement(xmlButtonSubmitField);
    this.ButtonCheckAvailabilityField.LoadFromXMLElement(xmlButtonCheckAvailabilityField);
    this.FocusingField.LoadFromXMLElement(xmlFocusingField);
    this.Items = new Array();
    for (var i = 0; i < xmlAdditionalFields.length; i++)
    {
      this.Items[i] = new TkpmAdditionalField();
      this.Items[i].LoadFromXMLElement(xmlAdditionalFields[i]);
    }
    return true;
  },

  GetAllPasswordFieldsAutofilled: function()
  {
    for (var i = 0; i < this.Items.length; i++)
      if ((this.Items[i].AutofillValueType == kpm_atPassword) &&
          (!this.Items[i].Autofilled)
         )
      {
        return false;
      }
    return true;
  },

  GetAutofilledFieldsCount: function()
  {
    var Result = 0;
    for (var i = 0; i < this.Items.length; i++)
      if (this.Items[i].Autofilled)
        Result++;
    return Result;
  },

  IndexOf: function(DestAdditionalField, opExcludeFieldValueCompare)
  {
    for (var i = 0; i < this.Items.length; i++)
      if (this.Items[i].Compare(DestAdditionalField, opExcludeFieldValueCompare))
        return i;
    return -1;
  },
}

// Additional Fields Autofill --------------------------------------------------

function kpmAdditionalFillElement(ADocument, AForm, AElement, AAdditionalField,
  AAutofillInfo, AAdditionalFieldsManager, AAdditionalFieldsAutofillInfo)
{
  var Result = false;
  // init AdditionalField from HTMLElement
  AAdditionalField.InitializeFromHTMLElement(ADocument, AForm, AElement);
  if (AAdditionalField.FieldType == kpm_afUnknown)
    return false;
  // ignore empty select element
  if (kpmIsSelectAdditionalFieldTypes(AAdditionalField.FieldType) &&
      (AElement.length == 0))
    return false;
  // try to find focusing element
  if ((AAdditionalFieldsAutofillInfo.FocusingElement == null) &&
      AAdditionalFieldsManager.FocusingField.Compare(AAdditionalField, 1))
  {
    AAdditionalFieldsAutofillInfo.FocusingElement = AElement;
    Result = true;
  } else    
  // try to find the submit button
  if ((AAdditionalField.FieldType == kpm_afButton) ||
      (AAdditionalField.FieldType == kpm_afButtonImage) ||
      (AAdditionalField.FieldType == kpm_afButtonSubmit)
     )
  {
    if ((AAdditionalFieldsAutofillInfo.SubmitButtonElement == null) &&
        AAdditionalFieldsManager.ButtonSubmitField.Compare(AAdditionalField, 1))
    {
      AAdditionalFieldsAutofillInfo.SubmitButtonElement = AElement;
      Result = true;
    } else
    if ((AAdditionalFieldsAutofillInfo.CheckAvailabilityButtonElement == null) &&
        AAdditionalFieldsManager.ButtonCheckAvailabilityField.Compare(AAdditionalField, 1))
    {
      AAdditionalFieldsAutofillInfo.CheckAvailabilityButtonElement = AElement;
      Result = true;
    } else    
    if (AAdditionalFieldsAutofillInfo.MustFindFocusingButtonElement)
    {
      if (AAdditionalField.FieldType == kpm_afButton)
      {
        if (AAdditionalFieldsAutofillInfo.FocusingButtonElement == null)
          AAdditionalFieldsAutofillInfo.FocusingButtonElement = AElement;
      } else
      if (AAdditionalField.FieldType == kpm_afButtonImage)
      {
        if (AAdditionalFieldsAutofillInfo.FocusingButtonImageElement == null)
          AAdditionalFieldsAutofillInfo.FocusingButtonImageElement = AElement;
      } else
      if (AAdditionalField.FieldType == kpm_afButtonSubmit)
      {
        if (AAdditionalFieldsAutofillInfo.FocusingButtonSubmitElement == null)
          AAdditionalFieldsAutofillInfo.FocusingButtonSubmitElement = AElement;
      }
    }
    return false;
  }
  var Index = AAdditionalFieldsManager.IndexOf(AAdditionalField,
    AAdditionalField.FieldType != kpm_afRadioButton);
  if (Index != -1)
  {
    Result = true;
    var AdditionalField = AAdditionalFieldsManager.Items[Index];
    // highlight element
    if (AAutofillInfo.AutofillReason == kpm_arHighlight || AAutofillInfo.AutofillReason == kpm_arAutofillAndHighlight)
      spHighlightHtmlElement(AElement);
    if (AAutofillInfo.AutofillReason != kpm_arAutofill && AAutofillInfo.AutofillReason != kpm_arAutofillAndHighlight)
      return Result;
    // fill element
    if (AdditionalField.AutofillValueType == kpm_atLogin)
    {
      // INFO: Force set focus to login element on facebook.com cause login cleared      
      if (kpmWideSearchText('facebook.com', AdditionalField.HtmlDocumentUrl) &&
          kpmCompareAdditionalFieldIDorName(AElement, 'email')
		 )
      {
        kpmHTMLElementSetFocus(AElement);
      }
      AdditionalField.FieldValue = AAutofillInfo.RealLogin.value; 
    } 
    else if (AdditionalField.AutofillValueType == kpm_atPassword)
    {
      AdditionalField.FieldValue = AAutofillInfo.Password.value;
    }
    AdditionalField.SetElementValue(AElement);
    if (AdditionalField.AutofillValueType == kpm_atLogin)
      AAutofillInfo.LoginAutofilled = AdditionalField.Autofilled;
    if (AdditionalField.Autofilled &&
        ((AdditionalField.AutofillValueType == kpm_atLogin) ||
         (AdditionalField.AutofillValueType == kpm_atPassword)
        )
       )
    {
      AAdditionalFieldsAutofillInfo.MustFindFocusingButtonElement = true;
    }
  }
  return Result;
}

function kpmAdditionalFillWindow(AWindow, AAutofillInfo, AAdditionalFieldsManager)
{
  var Result = true;
  if (AWindow && AAutofillInfo && AAdditionalFieldsManager)
  try
  {
    kpmAdditionalFieldsAutofillInfo = new TkpmAdditionalFieldsAutofillInfo();
    var AdditionalField = new TkpmAdditionalField();
    var DocumentList = kpmGetHTMLDocumentList(AWindow);
    for (var i = 0; i < DocumentList.length; i++)
    {
      var Document = DocumentList[i];      
      var Document2Autofilled = false;
      // try to autofill the Document's form
      if (Document.forms && Document.forms.length)
      {
        var DocumentsInputElements = Document.getElementsByTagName('input');
        for(var j = 0; j < Document.forms.length; j++)
        {
          var FormElement = Document.forms[j];
          AdditionalField.InitializeFromHTMLElement(Document, FormElement, FormElement);
          // is it FormSubmitField?
          if ((kpmAdditionalFieldsAutofillInfo.SubmitFormElement == null) &&
              AAdditionalFieldsManager.FormSubmitField.Compare(AdditionalField)
             )
          {
            kpmAdditionalFieldsAutofillInfo.SubmitFormElement = FormElement;
          }
          kpmAdditionalFieldsAutofillInfo.MustFindFocusingButtonElement = 0;
          // try to search form's elements
          var ElementList = kpmGetHTMLElementList(FormElement, DocumentsInputElements);
          for(var k = 0; k < ElementList.length; k++)
          {
            var Element = ElementList[k];
            if (kpmAdditionalFillElement(Document, FormElement, Element, AdditionalField,
              AAutofillInfo, AAdditionalFieldsManager, kpmAdditionalFieldsAutofillInfo))
            {
              Document2Autofilled = true;
            }
          }
        }
      }
      // if none forms found - try to aufotill Document elements without forms
      if (!Document2Autofilled/* &&
          (!Document.forms ||
           Document.forms.length == 0
          )*/
         )
      {
        var ElementList = kpmGetHTMLDocumentElementList(Document);
        for(var k = 0; k < ElementList.length; k++)
        {
          var Element = ElementList[k];
          kpmAdditionalFillElement(Document, null, Element, AdditionalField,
            AAutofillInfo, AAdditionalFieldsManager, kpmAdditionalFieldsAutofillInfo);
        }
      } 
    }
    if (AAutofillInfo.AutofillReason == kpm_arAutofill || AAutofillInfo.AutofillReason == kpm_arAutofillAndHighlight)
    {
      // click on 'Check availability' button
      kpmHTMLElementClick(kpmAdditionalFieldsAutofillInfo.CheckAvailabilityButtonElement);
      // set the focus on focusing element
      if (kpmAdditionalFieldsAutofillInfo.FocusingElement)
        kpmHTMLElementSetFocus(kpmAdditionalFieldsAutofillInfo.FocusingElement); else
      // set the focus on submit button or found focusing button
      if (AAutofillInfo.AllowSetFocus)
      {
        if (!kpmHTMLElementSetFocus(kpmAdditionalFieldsAutofillInfo.SubmitButtonElement))
          if (!kpmHTMLElementSetFocus(kpmAdditionalFieldsAutofillInfo.FocusingButtonSubmitElement))
            if (!kpmHTMLElementSetFocus(kpmAdditionalFieldsAutofillInfo.FocusingButtonImageElement))
              kpmHTMLElementSetFocus(kpmAdditionalFieldsAutofillInfo.FocusingButtonElement);
      }
      // autosubmit the form if need
      if (AAutofillInfo.AutoSubmit)
      {
        // first try to click the submit button, then try to submit the form
        if (!kpmHTMLElementClick(kpmAdditionalFieldsAutofillInfo.SubmitButtonElement))
          if (!kpmHTMLFormSubmit(kpmAdditionalFieldsAutofillInfo.SubmitFormElement, true))
            kpmLog.logError('kpmAdditionalFillWindow() kpmHTMLFormSubmit() == False ');

      }
      // if we doesn't fill any Password field or none fields at all -
      // return False to place password into clipboard
      if ((!AAdditionalFieldsManager.GetAllPasswordFieldsAutofilled()) ||
          (AAdditionalFieldsManager.GetAutofilledFieldsCount() == 0)
         )
      {
        Result = false;
      }
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmAdditionalFillWindow() Error: ' + ErrorMessage);
  }
  return Result;
}

// Password Control Exists Engine ----------------------------------------------

function kpmDocumentElementsPasswordControlExists(ADocument, AForm, AElementList, APasswordControlExistsInfo)
{
  if (ADocument && AElementList && APasswordControlExistsInfo)
  try
  {
    // fix focusing edit login/password
    kpmFocusingEditPasswordManager.FixFocusingEditList(AElementList);
    var InputPasswordElement = null;
    for(var i = 0; i < AElementList.length; i++)
    {
      var Element = AElementList[i];
      var IsLoginElement = {};
      // is it multipage authorization login?
      if (APasswordControlExistsInfo.PasswordControlMustBeFilled &&
          kpmWideSameText(Element.type, 'hidden'))
      {
        kpmMultipageAuthorizationManager.IsMultipageAuthorizationLogin(AForm,
          AElementList, Element, APasswordControlExistsInfo.PasswordControlMustBeFilled,
          IsLoginElement, APasswordControlExistsInfo.MultipageAuthorizationLogin
        );
      }
      // get element's additional field type
      var FieldType = kpmGetAdditionalFieldType(Element);
      if (kpmDebug_PasswordControlExists && kpmDebug_Elements && FieldType != kpm_afUnknown)
        kpmLog.logMessage('kpmDocumentElementsPasswordControlExists() FormElementList[' + (i+1) + '/' + AElementList.length + ']: <' +
          Element.tagName + '> <' + kpmGetElementName(Element) + '/' + Element.id + '>, FieldType: ' + kpmAdditionalFieldTypeToStr(FieldType) + ', ' +
          'RealFieldType: ' + kpmAdditionalFieldTypeToStr(kpmGetRealAdditionalFieldType(Element)) + ', ' +
          'Value: <' + Element.value + '>'
        );
      // ignore disallowed additional field elements
      if (!kpmIsAllowedAdditionalFieldElement(Element, FieldType, kpm_afAll))
      {
        if (kpmDebug_PasswordControlExists && kpmDebug_Elements)
          kpmLog.logMessage('kpmDocumentElementsPasswordControlExists() FormElementList[' + (i+1) + '/' + AElementList.length + ']: <' +
            Element.tagName + '> <' + kpmGetElementName(Element) + '/' + Element.id + '>, disallowed additional field element'
          );
        continue;
      }
      if (FieldType == kpm_afEdit)
      {
        // is it multipage authorization login?
        kpmMultipageAuthorizationManager.IsMultipageAuthorizationLogin(AForm,
          AElementList, Element, APasswordControlExistsInfo.PasswordControlMustBeFilled,
          IsLoginElement, APasswordControlExistsInfo.MultipageAuthorizationLogin
        );
        // obtain multipage login control rect
        if (!APasswordControlExistsInfo.PasswordControlMustBeFilled &&
            IsLoginElement.value &&
            kpmIsRectEmpty(APasswordControlExistsInfo.MultipageAuthorizationLoginControlRect)
           )
        {
          APasswordControlExistsInfo.MultipageAuthorizationLoginControlRect =
            kpmGetElementScreenPosition(Element);
        }
      } else
      if (FieldType == kpm_afEditPassword)
      {
        if (!InputPasswordElement)
        {
          // check edit password value
          if (APasswordControlExistsInfo.PasswordControlMustBeFilled)
          {
            // check if password specified?
            var Password = kpmGetAdditionalFieldValue(Element);
            if (kpmWideIsEmpty(Password))
            {
              // we can have case when in one form located more than one password controls some of them can be empty
              // for example: AJAX use only one form per document
              if (kpmPasswordControlLastValue.isEmpty(Element))
              {
                // try to find next non-empty password edit
                continue;
              }
            }
            // check if password is ignored hint?
            if (kpmAutofill().IsIgnoredPasswordHint(ADocument.URL, Password))
              continue;
          }
          // we found Password element
          InputPasswordElement = Element;
        }
      }
      // are we found password element?
      if (InputPasswordElement)
        break;
    }
    // are we found the password element?
    if (InputPasswordElement)
    {
      var Password = kpmGetAdditionalFieldValue(InputPasswordElement);
      if (kpmWideIsEmpty(Password))
        Password = kpmPasswordControlLastValue.get(InputPasswordElement);
      if (!APasswordControlExistsInfo.PasswordControlMustBeFilled ||
          !kpmWideIsEmpty(Password)
         )
      {
        // get form's action & method
        if (AForm)
        {
          APasswordControlExistsInfo.FormAction.value = kpmGetAdditionalFieldValue(AForm);
          APasswordControlExistsInfo.FormMethod = kpmGetHTMLFormMethod(AForm);
        } else
        {
          APasswordControlExistsInfo.FormAction.value = kpmGetDocumentUrl(ADocument)
          APasswordControlExistsInfo.FormMethod = kpm_htmlFormMethod_Post;
        }
        // get the password
        APasswordControlExistsInfo.Password.value = Password;
        // get the password control rect & empty state
        var tmpElement = InputPasswordElement;
	    while (tmpElement)
		{
		  APasswordControlExistsInfo.PasswordControlRect = kpmGetElementScreenPosition(tmpElement);
		  if (!kpmIsRectEmpty(APasswordControlExistsInfo.PasswordControlRect)) 
		    break;
		  tmpElement = tmpElement.parentNode;		  
		}
        APasswordControlExistsInfo.PasswordControlEmpty = kpmWideIsEmpty(APasswordControlExistsInfo.Password);
        return true;
      }
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmDocumentElementsPasswordControlExists() Error: ' + ErrorMessage);
  }
  return false;
}
  
function kpmWindowPasswordControlExists(AWindow, APasswordControlExistsInfo)
{
  if (AWindow)
  try
  {
    var DocumentList = kpmGetHTMLDocumentList(AWindow);
    for (var i = 0; i < DocumentList.length; i++)
    {
      var Document = DocumentList[i];
      if (Document.forms && Document.forms.length)
      {
        var DocumentsInputElements = Document.getElementsByTagName('input');
        for(var j = 0; j < Document.forms.length; j++)
        {
          var FormElement = Document.forms[j];
          if (kpmDebug_PasswordControlExists && kpmDebug_Elements)
            kpmLog.logMessage('kpmWindowPasswordControlExists() Forms[' + (j+1) + '/' + Document.forms.length + ']: ' + kpmGetAdditionalFieldName(FormElement));
          // check if form is allowed?
          if (!kpmIsAllowedAdditionalFieldForm(FormElement))
          {
            if (kpmDebug_PasswordControlExists && kpmDebug_Elements)
              kpmLog.logMessage('kpmWindowPasswordControlExists() NOT is allowed form!');
            continue;
          }
          // enum the elements of FormElement
          var ElementList = kpmGetHTMLElementList(FormElement, DocumentsInputElements);
          if (kpmDocumentElementsPasswordControlExists(Document, FormElement, ElementList, APasswordControlExistsInfo))
            return true;
        }
      }
      // if none forms found - try to find password control without forms
      if (!Document.forms ||
          Document.forms.length == 0
         )
      {
        var ElementList = kpmGetHTMLDocumentElementList(Document);
        if (kpmDocumentElementsPasswordControlExists(Document, null, ElementList, APasswordControlExistsInfo))
          return true;        
      }
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmWindowPasswordControlExists() Error: ' + ErrorMessage);
  }
  return false; 
}

function kpmPasswordControlExists(AWindow, APasswordControlExistsInfo)
{
  APasswordControlExistsInfo.PasswordControlMustBeFilled = false;
  var Result = kpmWindowPasswordControlExists(AWindow, APasswordControlExistsInfo);
  // get password control pos from multipage login one
  if (kpmIsRectEmpty(APasswordControlExistsInfo.PasswordControlRect))
    APasswordControlExistsInfo.PasswordControlRect = APasswordControlExistsInfo.MultipageAuthorizationLoginControlRect;
  return Result;
}

function kpmNotEmptyPasswordControlExists(AWindow, APasswordControlExistsInfo)
{
  APasswordControlExistsInfo.PasswordControlMustBeFilled = true;
  return kpmWindowPasswordControlExists(AWindow, APasswordControlExistsInfo) && 
    (!APasswordControlExistsInfo.PasswordControlEmpty);
}

function kpmNotEmptyPasswordControlExistsInForm(AWindow, AForm, APasswordControlExistsInfo)
{
  APasswordControlExistsInfo.PasswordControlMustBeFilled = true;
  try
  {
    var InputPasswordElement = null;
    var FormElement = AForm;
    if (kpmDebug_NotEmptyPasswordControlExistsInForm && kpmDebug_Elements)
      kpmLog.logMessage('kpmNotEmptyPasswordControlExistsInForm() FormElement: ' + kpmGetAdditionalFieldName(FormElement));
    if (kpmIsAllowedAdditionalFieldForm(FormElement))
    {
      var ADocument = FormElement.ownerDocument;
      // enum the elements of FormElement
      for(var j = 0; j < FormElement.elements.length; j++)
      {
        var Element = FormElement.elements[j];
        var IsLoginElement = {};
        // is it multipage authorization login?
        if (kpmWideSameText(Element.type, 'hidden'))
        {
          kpmMultipageAuthorizationManager.IsMultipageAuthorizationLogin(FormElement,
            FormElement.elements, Element, APasswordControlExistsInfo.PasswordControlMustBeFilled,
            IsLoginElement, APasswordControlExistsInfo.MultipageAuthorizationLogin
          );
        }
        var FieldType = kpmGetAdditionalFieldType(Element);
        if (kpmDebug_NotEmptyPasswordControlExistsInForm && kpmDebug_Elements && FieldType != kpm_afUnknown)
          kpmLog.logMessage('kpmNotEmptyPasswordControlExistsInForm() FormElementList[' + (j+1) + '/' + FormElement.elements.length + ']: <' +
            Element.tagName + '> <' + kpmGetElementName(Element) + '/' + Element.id + '>, FieldType: ' + kpmAdditionalFieldTypeToStr(FieldType) + ', ' +
            'RealFieldType: ' + kpmAdditionalFieldTypeToStr(kpmGetRealAdditionalFieldType(Element)) + ', ' +
            'Value: <' + Element.value + '>'
          );
        // ignore disallowed additional field elements
        if (!kpmIsAllowedAdditionalFieldElement(Element, FieldType, kpm_afAll))
          continue;
        if (FieldType == kpm_afEdit)
        {
          // is it multipage authorization login?
          kpmMultipageAuthorizationManager.IsMultipageAuthorizationLogin(FormElement,
            FormElement.elements, Element, APasswordControlExistsInfo.PasswordControlMustBeFilled,
            IsLoginElement, APasswordControlExistsInfo.MultipageAuthorizationLogin
          );
        } else
        if (FieldType == kpm_afEditPassword)
        {
          if (InputPasswordElement == null)
          {
            // check if password specified?
            var Password = kpmGetAdditionalFieldValue(Element);
            if (kpmWideIsEmpty(Password))
            {
              // we can have case when in one form located more than one password controls some of them can be empty
              // for example: AJAX use only one form per document
              // try to find next non-empty password edit
              continue;
            }
            // check if password is ignored hint?
            if (kpmAutofill().IsIgnoredPasswordHint(ADocument.URL, Password))
              continue;
            // we found Password element
            InputPasswordElement = Element;
          }
        }
        // must we break?
        if (InputPasswordElement != null)
          break;
      }
    }
    // are we found password control?
    if (InputPasswordElement != null)
    {
      // get the password
      APasswordControlExistsInfo.Password.value = kpmGetAdditionalFieldValue(InputPasswordElement);
    }
    // if password isn't found - try to search trough all the window forms
    if (kpmWideIsEmpty(APasswordControlExistsInfo.Password))
    {
      // search trough all the window forms
      APasswordControlExistsInfo.PasswordControlEmpty = true;
      kpmNotEmptyPasswordControlExists(AWindow, APasswordControlExistsInfo);
    } 
    // finalize found values
    APasswordControlExistsInfo.FormAction.value = kpmGetAdditionalFieldValue(AForm);
    APasswordControlExistsInfo.FormMethod = kpmGetHTMLFormMethod(AForm);
    APasswordControlExistsInfo.PasswordControlEmpty = kpmWideIsEmpty(APasswordControlExistsInfo.Password);
    if (kpmDebug_NotEmptyPasswordControlExistsInForm)
    {
      kpmLog.logMessage('kpmNotEmptyPasswordControlExistsInForm(), ' + 
        'Password: <' + APasswordControlExistsInfo.Password.value + '>, ' + 
        'PasswordControlEmpty: ' + APasswordControlExistsInfo.PasswordControlEmpty + ', ' +
        'FormAction: <' + APasswordControlExistsInfo.FormAction.value + '>,' +
        'FormMethod: <' + APasswordControlExistsInfo.FormMethod + '>');
    }
    return !APasswordControlExistsInfo.PasswordControlEmpty;
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmNotEmptyPasswordControlExistsInForm() Error: ' + ErrorMessage);
    return false;
  }
}

// Export DOM to Xml engine ----------------------------------------------------

function kpmExportDOMForm(ADocument, AForm, AXmlDocument, AFormsNode)
{
  var FormNode = null;
  if (AXmlDocument && AFormsNode)
  try
  {
    // create Form node
    FormNode = AXmlDocument.createElement('form');
    AFormsNode.appendChild(FormNode);
    var s;
    if (AForm)
    {
      // id
      if (!kpmWideIsEmpty(AForm.id))
        FormNode.setAttribute('id', AForm.id);
      // name
      s = kpmGetAdditionalFieldName(AForm);
      if (!kpmWideIsEmpty(s))
        FormNode.setAttribute('name', s);
      // action
      s = kpmGetAdditionalFieldValue(AForm);
    } else
    {
      // action
      s = kpmGetDocumentUrl(ADocument);
    }
    s = kpmWidePrepareLowerCaseStr(s);
    if (!kpmWideIsEmpty(s))
      FormNode.setAttribute('action', s);
    // documentUrl
    s = kpmWidePrepareLowerCaseStr(kpmGetDocumentUrl(ADocument));
    if (!kpmWideIsEmpty(s))
      FormNode.setAttribute('documentUrl', s);      
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmExportDOMForm() Error: ' + ErrorMessage);
  }
  return FormNode;
}

function kpmExportDOMElement(AElement, AFieldType, AXmlDocument, AFocusedElementName, AFormNode)
{
  var s;
  var ElementNode = AXmlDocument.createElement('element');
  AFormNode.appendChild(ElementNode);
  if (kpmIsInputAdditionalFieldTypes(AFieldType))
    ElementNode.setAttribute('type', 'input'); else
  if (kpmIsTextAreaAdditionalFieldTypes(AFieldType))
    ElementNode.setAttribute('type', 'textarea'); else
  if (kpmIsSelectAdditionalFieldTypes(AFieldType))
    ElementNode.setAttribute('type', 'select');
  if (!kpmWideIsEmpty(AElement.id))
    ElementNode.setAttribute('id', AElement.id);
  s = kpmGetAdditionalFieldName(AElement);
  if (!kpmWideIsEmpty(s))
  {
    ElementNode.setAttribute('name', s);
    if (kpmWideSameText(s, AFocusedElementName))
      ElementNode.setAttribute('focused', true);
  }
  if (kpmIsInputAdditionalFieldTypes(AFieldType))
  {
    var ucTagName = '';
    if (AElement.tagName)
      ucTagName = kpmWidePrepareUpperCaseStr(AElement.tagName);
    // ButtonElement or InputElement
    if (ucTagName == kpm_szBUTTON ||
        ucTagName == kpm_szINPUT
       )
    {
      var lcType = kpmWidePrepareLowerCaseStr(AElement.type);
      // on amazon.com text field has 'email' type, so change it to 'text' one!
      if (kpmWideSameText(lcType, 'email'))
        lcType = 'text';
      ElementNode.setAttribute('inputtype', lcType);
    } else
    if (AFieldType == kpm_afButtonSubmit)
    {
      // this is for case <a>/<img> tags
      ElementNode.setAttribute('inputtype', 'submit');
    }
  }
  if (((AFieldType == kpm_afEdit) ||
       (AFieldType == kpm_afEditPassword)
      ) &&
      (AElement.maxLength > 0)
     )
  {
    ElementNode.setAttribute('maxLength', AElement.maxLength);
  }
  s = kpmGetAdditionalFieldValue(AElement);
  if (!kpmWideIsEmpty(s))
    ElementNode.setAttribute('value', s);
  // onclick
  if (AFieldType == kpm_afButton || AFieldType == kpm_afButtonImage || AFieldType == kpm_afButtonSubmit)
  {
    var funAttribute = AElement.getAttribute('onclick');
    if (funAttribute)
      s = funAttribute.toString();
    else
      s = '';
    if (!kpmWideIsEmpty(s))
      ElementNode.setAttribute('onclick', s);
  }
  if (kpmIsSelectAdditionalFieldTypes(AFieldType))
  {
    // Element's options
    for(var k = 0; k < AElement.length; k++)
    {
      var Option = AElement.options[k];
      var OptionNode = AXmlDocument.createElement('item');
      ElementNode.appendChild(OptionNode);
      if (!kpmWideIsEmpty(Option.text))
        OptionNode.setAttribute('name', Option.text);
      if (!kpmWideIsEmpty(Option.value))
        OptionNode.setAttribute('value', Option.value);
    }
  }
}

function kpmExportDOMElementsToXml(ADocument, AForm, AElementList, AXmlDocument, AFocusedElementName, AFormsNode)
{
  var Result = false;
  try
  {
    var FormNode = null;
    // fix focusing edit login/password
    kpmFocusingEditPasswordManager.FixFocusingEditList(AElementList);
    for(var i = 0; i < AElementList.length; i++)
    {
      var Element = AElementList[i];
      var FieldType = kpmGetAdditionalFieldType(Element);
      // ignore disallowed additional field elements
      if (!kpmIsAllowedAdditionalFieldElement(Element, FieldType, kpm_afAll))
        continue;
      // ignore empty select element
      if (kpmIsSelectAdditionalFieldTypes(FieldType) && (Element.length == 0))
        continue;
      // export Form
      if (FormNode == null)
        FormNode = kpmExportDOMForm(ADocument, AForm, AXmlDocument, AFormsNode);
      // log element if need
      if (kpmDebug_ExportDOMToXml && kpmDebug_Elements)
        kpmLog.logMessage('kpmExportDOMElementsToXml() FormElementList[' + (i+1) + '/' + AElementList.length + ']: <' +
          Element.tagName + '> <' + kpmGetElementName(Element) + '/' + Element.id + '>, FieldType: ' + kpmAdditionalFieldTypeToStr(FieldType) + ', ' +
          'RealFieldType: ' + kpmAdditionalFieldTypeToStr(kpmGetRealAdditionalFieldType(Element)) + ', ' +
          'Value: <' + Element.value + '>'
        );
      // export Element
      kpmExportDOMElement(Element, FieldType, AXmlDocument, AFocusedElementName, FormNode);
      Result = true;
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmExportDOMElementsToXml() Error: ' + ErrorMessage);
    Result = false;
  }
  return Result;
}

function kpmExportDOMToXml(AWindow)
{
  var xml = '';
  if (AWindow)
  try
  {
    // create XmlDocument
    var XmlDocument = document.implementation.createDocument('', 'forms', null);
    XmlDocument.async = false;
    var FormsNode = XmlDocument.documentElement;
    // export DOM from AWindow
    var DocumentList = kpmGetHTMLDocumentList(AWindow);
    if (kpmDebug_ExportDOMToXml && kpmDebug_Elements)
      kpmLog.logMessage('kpmExportDOMToXml() DocumentList.length=' + DocumentList.length);
    var FocusedElementName = kpmGetAdditionalFieldName(kpmGetDocumentFocusedElement(AWindow.document));
    for (var i = 0; i < DocumentList.length; i++)
    {
      var Document = DocumentList[i];
      var Document2Exported = false;
      if (kpmDebug_ExportDOMToXml && kpmDebug_Elements)
        kpmLog.logMessage('kpmExportDOMToXml() Documents[' + (i+1) + '/' + DocumentList.length + ']: ' + 
          kpmGetDocumentUrl(Document) + ' Document.forms.length=' + Document.forms.length);
      // enum forms
      if (Document.forms && Document.forms.length)
      {
        var DocumentsInputElements = Document.getElementsByTagName('input');
        for(var j = 0; j < Document.forms.length; j++)
        {
          var FormElement = Document.forms[j];
          if (kpmDebug_ExportDOMToXml && kpmDebug_Elements)
            kpmLog.logMessage('kpmExportDOMToXml() Forms[' + (j+1) + '/' + Document.forms.length + ']: ' + 
              kpmGetAdditionalFieldName(FormElement));
          // check if form is allowed?
          if (!kpmIsAllowedAdditionalFieldForm(FormElement))
          {
            if (kpmDebug_ExportDOMToXml && kpmDebug_Elements)
              kpmLog.logMessage('kpmExportDOMToXml() NOT is allowed form!');
            continue;
          }
          var FormNode = null;
          var ElementList = kpmGetHTMLElementList(FormElement, DocumentsInputElements);
          if (kpmExportDOMElementsToXml(Document, FormElement, ElementList, XmlDocument, FocusedElementName, FormsNode))
          {
            Document2Exported = true;
          }
        }
      }
      // if none forms found - try to export Document elements without forms
      if (!Document2Exported/* &&
          (!Document.forms ||
           Document.forms.length == 0          
          )*/
         )
      {
        if (kpmDebug_ExportDOMToXml && kpmDebug_Elements)
          kpmLog.logError('kpmExportDOMToXml() none forms found');
        var ElementList = kpmGetHTMLDocumentElementList(Document);
        kpmExportDOMElementsToXml(Document, null, ElementList, XmlDocument, FocusedElementName, FormsNode);
      }
    }
    // check if any DOM element exported
    if (FormsNode.hasChildNodes())
    {
      // serialize xml from XmlDocument
      var XmlSerializer = kpmNewXMLSerializer();
      xml = '<?xml version="1.0" encoding="UTF-16"?>';
      if (kpmDebug_ExportDOMToXml)
      {
        var s = '';
        try
        {
          s = XML(XmlSerializer.serializeToString(XmlDocument)).toXMLString();
        }
        catch (ErrorMessage)
        {
          s = XmlSerializer.serializeToString(XmlDocument);
        }
        kpmLog.logMessage('kpmExportDOMToXml() xml:' + kpmLineBreak + xml + s);
      }
      xml = xml + XmlSerializer.serializeToString(XmlDocument);
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmExportDOMToXml() Error: ' + ErrorMessage);
  }
  return xml;
}

// Event Handlers --------------------------------------------------------------

function kpmAutofillDocument(AWindow, IsOnLoad)
{
  if (kpmAutofill())
  try
  {
    // enable cross-domains frames access
    try
    {
      netscape.security.PrivilegeManager.enablePrivilege('UniversalBrowserRead');
    }
    catch (ErrorMessage)
    {
      // keep silence
    }

    var objIdentityFormIndex = {};
    var objAutofillReason = {};
    var objIdentityAutofilling = {};
    var strAccountID = {};
    var strAccountName = {};
    var kpmAutofillInfo = new TkpmAutofillInfo();
    var strAdditionalFieldsXml = {};
    var objUrlIsAccountLink = {};
    var objAllowSetFocus = {};
    var objAutoSubmit = {};

    if (IsOnLoad)
    {
      // document is complete      
      kpmSetIsDocumentComplete(AWindow.document);
      // DocumentComplete()
      var IsActiveTab = kpmIsActiveTabWindow(AWindow);
      // WARNING: We must always call szPasswordControlExists() to obtain PasswordControlExists & FormAction
      var kpmPasswordControlExistsInfo = new TkpmPasswordControlExistsInfo();
      var PasswordControlExists = kpmPasswordControlExists(AWindow, kpmPasswordControlExistsInfo);
      // try to export DOM to xml
      var DOMXml = kpmExportDOMToXml(AWindow);
      if (kpmWideIsEmpty(DOMXml))
      {
        if (kpmDebug_Autofill)
          kpmLog.logMessage('kpmAutofillDocument() DOMXml is empty!');
        return;
      }
      if (!kpmAutofill().DocumentComplete(
          // in
          spAuthKey, kpmWindowContent(window).document.title,
          kpmGetDocumentUrl(AWindow.document), AWindow.document.title,
          DOMXml, IsActiveTab, PasswordControlExists,
          kpmPasswordControlExistsInfo.PasswordControlRect.left,
          kpmPasswordControlExistsInfo.PasswordControlRect.top,
          kpmPasswordControlExistsInfo.PasswordControlRect.right,
          kpmPasswordControlExistsInfo.PasswordControlRect.bottom,
          kpmPasswordControlExistsInfo.FormAction.value,
          kpmPasswordControlExistsInfo.FormMethod,
          // out
          objIdentityFormIndex, objAutofillReason,
          strAccountID, strAccountName, kpmAutofillInfo.Login, kpmAutofillInfo.RealLogin, 
          kpmAutofillInfo.Password, strAdditionalFieldsXml,
          objUrlIsAccountLink, objAllowSetFocus, objAutoSubmit)
         )
      {
        if (kpmDebug_Autofill)
          kpmLog.logMessage('kpmAutofillDocument() Not kpmAutofill().DocumentComplete()!');
        return;
      }
    } else
    {
      // ManualAutofill()
      if (!kpmAutofill().ManualAutofill(
          // in
          spAuthKey,
          // out
          objIdentityAutofilling, objIdentityFormIndex, objAutofillReason,
          strAccountID, strAccountName, kpmAutofillInfo.Login, kpmAutofillInfo.RealLogin,
          kpmAutofillInfo.Password, strAdditionalFieldsXml,
          objUrlIsAccountLink, objAllowSetFocus, objAutoSubmit)
         )
      {
        if (kpmDebug_Autofill)
          kpmLog.logMessage('kpmAutofillDocument() Not kpmAutofill().ManualAutofill()!');
        return;
      }
      
      // reset last focused element - to allow show Password Generator balloon
      //   if edit password is focused during identity autofill
      if (objIdentityAutofilling.value)
      {
        kpmLastFocusedElement = null;
      }
    }
    kpmAutofillInfo.AllowSetFocus = objAllowSetFocus.value;
    kpmAutofillInfo.AutoSubmit = objAutoSubmit.value;
    kpmAutofillInfo.IdentityFormIndex = objIdentityFormIndex.value;
    kpmAutofillInfo.AutofillReason = objAutofillReason.value;

    kpmIsAutofillingDocument = true;
    try
    {
      if (kpmDebug_Autofill)
        kpmLog.logMessage('kpmAutofillDocument() Before autofilling document. ' + 
          'AccountName: <' + strAccountName.value + '>, ' +
          'URL: <' + kpmGetDocumentUrl(AWindow.document) + '>, ' +
          'AdditionalFieldsXml: "' + strAdditionalFieldsXml.value + '"'
        );
      // create AdditionalFieldsManager and load from external file
      var AdditionalFieldsManager = new TkpmAdditionalFieldsManager();
      if (AdditionalFieldsManager.LoadFromXml(strAdditionalFieldsXml.value))
        kpmAutofillInfo.PasswordAutofilled = kpmAdditionalFillWindow(AWindow, kpmAutofillInfo, AdditionalFieldsManager);
    }
    catch (ErrorMessage)
    {
      kpmLog.logError('kpmAutofillDocument() Error while autofilling document: ' + ErrorMessage);
    }
    kpmIsAutofillingDocument = false;

    if (kpmAutofillInfo.AutofillReason == kpm_arAutofill || kpmAutofillInfo.AutofillReason == kpm_arAutofillAndHighlight)
    {
      // callback the Password Manager to display login information or
      // to place password to clipboard (only if user manual activate password)
      if (kpmAutofillInfo.PasswordAutofilled || !IsOnLoad)
      {
        kpmAutofill().AutofillCallback(kpmWindowContent(window).document.title,
          kpmAutofillInfo.LoginAutofilled, kpmAutofillInfo.PasswordAutofilled,
          objIdentityAutofilling.value,
          strAccountID.value, strAccountName.value, kpmAutofillInfo.Login.value);
      }
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmAutofillDocument() Error: ' + ErrorMessage);
  }      
}

function kpmAllowProceedDocument(ADocument, AllowAboutBlank)
{
  try
  {
    // check for kpmAutofill & ADocument objects initialized
    if (kpmAutofill() == null || ADocument == null)
      return false;
    // filter:
    //  * empty URLs
    var Url = kpmGetDocumentUrl(ADocument);
    if (kpmWideIsEmpty(Url))
      return false;
    //  * non-document nodes
    if (ADocument.nodeName != '#document')
      return false;
    //  * disallow XML Content-Type
    if (kpmWideSearchText('XML', ADocument.contentType) &&
        !kpmWideSearchText('HTML', ADocument.contentType)
       )
    {
      return false;
    }
    //  * blank URLs
    if (kpmWideSearchText('about:', Url) &&
        !AllowAboutBlank
       )
    {
      return false;
    }
    return true;
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmAllowProceedDocument() Error: ' + ErrorMessage);
    return false;
  }
}

// Event handlers

var TimeredAutofillDocument = null;

function kpmTimeredAutofillDocument(AEvent)
{
  // try to autofill document by timer
  if (TimeredAutofillDocument &&
      kpmWindowContent(window)
     )
  {
    try
    {
      var Event = new Object();
      Event.target = TimeredAutofillDocument;
      kpmOnLoad(Event);
    }
    catch(ErrorMessage)
    {
      kpmLog.logError('kpmTimeredAutofillDocument() Error: ' + ErrorMessage);
    }
    TimeredAutofillDocument = null;
  }
}

function kpmOnLoad(AEvent)
{
  if (AEvent)
  try
  {
    if (!kpmAdditionalEventsAdded)
    {
      kpmAdditionalEventsAdded = true;
      kpmAddAdditionalEvents();
    }
    var ADocument = null;
    if (kpmIsOpera())
    {
      ADocument = AEvent.target;
    }
    else if (AEvent.target instanceof HTMLDocument)
    {
      ADocument = AEvent.target;
    }
    else if (AEvent.target)
    {
      // AEvent.target can be the XPCNativeWrapper!
      ADocument = AEvent.target.wrappedJSObject;
    }
    if (kpmIsHTMLDocument(ADocument))
    {
      if (!kpmAllowProceedDocument(ADocument, true))
        return;
      // WARNING: Filter loaded event of child frames! 'window.parent = window' - the window parent equal itself if there is no parent
      var AWindow = ADocument.defaultView;
      // install HTMLFormElement prototype for window
      kpmInstallHTMLFormElementPrototypeForWindow(AWindow, false);
      // check for frame element
      if (AWindow.frameElement || AWindow != AWindow.parent)
        return;

      if (kpmDebug_OnLoad)
      {
        var IsActiveTab = 'false';
        if (kpmIsActiveTabWindow(AWindow))
          IsActiveTab = 'true';
        kpmLog.logMessage('kpmOnLoad() URL: ' + kpmGetDocumentUrl(AWindow.document) +
          ', ADocument.URL: ' + kpmGetDocumentUrl(ADocument) +
          ', frames.length: ' + AWindow.frames.length +
          ', IsActiveTab: ' + IsActiveTab
        );
      }
      // check the loaded page
      var Url = kpmWidePrepareLowerCaseStr(kpmGetDocumentUrl(ADocument));
      if (kpmProcessOnLoad(ADocument, Url))
        return;
      // use timered autofill for Url?
      if ((TimeredAutofillDocument != ADocument) &&
          kpmIsTimeredAutofillUrl(Url)
         )
      {
        TimeredAutofillDocument = ADocument;
        var interval = kpmGetTimeredAutofillUrlInterval(Url);
        setTimeout('kpmTimeredAutofillDocument()', interval);
        return;
      }
      // autofill window
      kpmAutofillDocument(AWindow, 1);
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmOnLoad() Error: ' + ErrorMessage);
  }
}

function kpmOnSubmit(AEvent)
{
  if (AEvent)
  try
  {
    var AForm = AEvent.target;
    if (AForm)
    {
      if (kpmWideSearchText('javascript:', AForm.action))
      {
        if (kpmDebug_OnSubmit)
          kpmLog.logMessage('kpmOnSubmit() Denied AForm.action: ' + AForm.action);
        return;
      }
      var ADocument = AForm.ownerDocument;
      // on ASP NET sites we can get window in "AEvent.target"
      if (!ADocument && AEvent.target.document) 
      {				  
        ADocument = AEvent.target.document;
		if (ADocument.forms && ADocument.forms.length)
		{
		  AForm = ADocument.forms[0];  
        } 
        else  
		{ 
		  kpmLog.logError('kpmOnSubmit() can not find form!');
		  return;
		}
      }
      if (!kpmAllowProceedDocument(ADocument, false))
      {
        kpmLog.logError('kpmOnSubmit() NOT kpmAllowProceedDocument()!');
        return;
      }
      if (kpmIsAutofillingDocument)
      {
        if (kpmDebug_OnSubmit)
          kpmLog.logError('kpmOnSubmit() kpmIsAutofillingDocument!');
        return;
      }
      var AWindow = ADocument.defaultView;
      // INFO: In Google Chrome there is a bug.
      //   If some form is located in the "iframe" element and it is submitted by JavaScript
      //   then we get "null" as "document.defaultView" here.
      if (!AWindow)
        AWindow = window;
      // convert AWindow to root window (cause of frames using)
      while (AWindow.frameElement || AWindow != AWindow.parent)
        AWindow = AWindow.parent;
      if (kpmDebug_OnSubmit)
      {
        kpmLog.logMessage('kpmOnSubmit() URL: ' + kpmGetDocumentUrl(AWindow.document) +
          ', ADocument.URL: ' + kpmGetDocumentUrl(ADocument) +
          ', frames.length: ' + AWindow.frames.length +
          ', AForm.action: ' + AForm.action);
      }
      kpmAutofill().HideShowedCaptionButtonBalloon();
      var kpmPasswordControlExistsInfo = new TkpmPasswordControlExistsInfo();
      if (kpmNotEmptyPasswordControlExistsInForm(AWindow, AForm, kpmPasswordControlExistsInfo))
      {
        var objCancel = {};
        var strBeforeSubmitManagerXml = {};
        var DOMXml = kpmExportDOMToXml(AWindow);
        kpmAutofill().BeforeSubmit(kpmGetDocumentUrl(AWindow.document), AWindow.document.title, 
          '', kpmPasswordControlExistsInfo.Password.value, DOMXml,
          kpmPasswordControlExistsInfo.FormAction.value, kpmPasswordControlExistsInfo.FormMethod, 
          objCancel, strBeforeSubmitManagerXml
        );
        if (objCancel.value)
        {
          // we must cancel submit cause of security reasons
          AEvent.preventDefault();
        } else
        if (!kpmWideIsEmpty(strBeforeSubmitManagerXml))
        {
          // create BeforeSubmitManager and load from Xml
          var BeforeSubmitManager = new TkpmBeforeSubmitManager();
          if (BeforeSubmitManager.LoadFromXML(strBeforeSubmitManagerXml.value))
            kpmFillBeforeSubmitWindow(AWindow, BeforeSubmitManager);
        }
      } else
      {
        // send multipage authorization info
        if (!kpmWideIsEmpty(kpmPasswordControlExistsInfo.MultipageAuthorizationLogin))
        {
          kpmAutofill().StoreMultipageAuthorizationInfo(kpmGetDocumentUrl(AWindow.document),
            kpmPasswordControlExistsInfo.MultipageAuthorizationLogin.value);
        }
      }
    } else
    {
      kpmLog.logError('kpmOnSubmit() Unknown AForm!');
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmOnSubmit() Error: ' + ErrorMessage);
  }
}

function kpmGetTickCount()
{
  var now = new Date();
  return now.getTime(); // milliseconds since 1970/01/01
}

var kpmLastOnSubmitTickCount = 0;

function kpmAllowOnSubmitByTickCount()
{
  if (kpmLastOnSubmitTickCount == 0)
    return true;
  else
    return kpmGetTickCount() - kpmLastOnSubmitTickCount > 3000; // in milliseconds
}

function kpmOnSubmit_Base(AEvent, SenderFunctionName)
{
  try
  {                         
    var AForm = AEvent.target;
    if (kpmDebug_OnSubmit)
    {
      var FormAction = kpmGetAdditionalFieldValue(AForm);
      var strOnSubmit = '';
      var funOnSubmit = AForm.getAttribute('onsubmit');
      if (funOnSubmit)
        strOnSubmit = funOnSubmit.toString();
      kpmLog.logMessage(SenderFunctionName + '(AForm) name/id: <' + AForm.name + '/' + AForm.id + '> ' +
        'action: <' + FormAction + '> onsubmit: <' + strOnSubmit + '>'
      );
    }
    
    // check if focused element isn't search element
    var AFormADocument = AForm.ownerDocument;
    var ActiveElement = kpmGetDocumentFocusedElement(AFormADocument); 
    if (ActiveElement) 
    {
      var FieldType = kpmGetAdditionalFieldType(ActiveElement);
      if (FieldType == kpm_afEdit && (!kpmIsAllowedAdditionalFieldElement(ActiveElement, FieldType, kpm_afAll))) 
        return false;
    }
       
    // check if we allow call kpmOnSubmit()?
    if (!kpmAllowOnSubmitByTickCount())
    {
      if (kpmDebug_OnSubmit)
        kpmLog.logMessage(SenderFunctionName + '(), DENY kpmOnSubmit()! by TickCount: kpmGetTickCount() - kpmLastOnSubmitTickCount=' + (kpmGetTickCount() - kpmLastOnSubmitTickCount));
      return false;
    }
    // call kpmOnSubmit()
    var Result = kpmOnSubmit(AEvent);
    kpmLastOnSubmitTickCount = kpmGetTickCount();
    return Result; 
  }
  catch (ErrorMessage)
  {
    kpmLog.logError(SenderFunctionName + '() Error: ' + ErrorMessage);
    return false;
  }
}
 
function kpmOnSubmit_Common(AEvent)
{
  return kpmOnSubmit_Base(AEvent, 'kpmOnSubmit_Common');
}

function kpmOnSubmit_JS(AEvent)
{
  return kpmOnSubmit_Base(AEvent, 'kpmOnSubmit_JS');
}

function kpmOnBeforeUnload()
{
  if (kpmAutofill() == null)
    return;

  kpmAutofill().HideShowedCaptionButtonBalloon();
}

function kpmOnUnload()
{
  if (kpmAdditionalEventsAdded)
    kpmRemoveAdditionalEvents();
  kpmRemoveBaseEvents();
  kpmStopAutofillEngineTimer();
}

function kpmOnEditPasswordChange(AEvent)
{
  try
  {
    // set PasswordControlLastValue
    var AEdit = AEvent.target;
    if (AEdit)
      kpmPasswordControlLastValue.set(AEdit);
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmOnEditPasswordChange() Error: ' + ErrorMessage);
  }
}

function kpmInstallHTMLFormElementPrototypeForWindow(AWindow, AOnChangeLocation)
{
  if (AWindow)
  try
  {
    var DocumentList = kpmGetHTMLDocumentList(AWindow);
    for (var i = 0; i < DocumentList.length; i++)
    {
      var ADocument = DocumentList[i];
      if (ADocument.wrappedJSObject)
        ADocument = ADocument.wrappedJSObject;
      var Url = kpmGetDocumentUrl(ADocument);
      if (!kpmAllowProceedDocument(ADocument, false))
      {
        if (kpmDebug_InstallHTMLFormElementPrototypeForDocument)
          kpmLog.logMessage('kpmInstallHTMLFormElementPrototypeForDocument(): Disallowed document for <' + Url + '>');
        continue;
      }
      if (AOnChangeLocation)
      {
        /*
        if (kpmWideSearchText('facebook.com', Url) ||
            kpmWideSearchText('avg.com', Url) ||
            kpmWideSearchText('seznam.cz', Url) ||
            kpmWideSearchText('download.com', Url) ||
            kpmWideSearchText('download.cnet.com', Url) ||
            kpmWideSearchText('compete.com', Url) ||
            kpmWideSearchText('microsoft.com', Url) ||
            kpmWideSearchText('nnm.ru', Url) ||
            kpmWideSearchText('qip.ru', Url) ||
            kpmWideSearchText('ironkey.com', Url) ||
            kpmWideSearchText('youtube.com', Url) ||
            kpmWideSearchText('alltel.com', Url)
           )
        */
        {
          return;
        }
      }
      // INFO: In Google Chrome there is a bug. 
      //   If some document is located in the "iframe" element 
      //   then we get "null" as "document.defaultView" here.
      //   So we must force add event listener to the document!
      if (kpmIsChrome())
      {
        if (!ADocument.defaultView)
        {
          // add event listener to the document
          ADocument.addEventListener('kpm_submit', kpmOnSubmit_JS, false, true);
          // add event listener to the forms of document
          for(var j = 0; j < ADocument.forms.length; j++)
          {
            var form = ADocument.forms[j];
            form.addEventListener('submit', kpmOnSubmit_Common, true);
          }
        }
      }
      var ParentElements = ADocument.getElementsByTagName('head');
      if (ParentElements && ParentElements.length)
      {
        var ParentElement = ParentElements[0];
        if (ParentElement)
        {
          if (ADocument.getElementById("objHTMLFormElementPrototype") == null)
          {
            if (kpmDebug_InstallHTMLFormElementPrototypeForDocument)
              kpmLog.logMessage('kpmInstallHTMLFormElementPrototypeForDocument(): Install objHTMLFormElementPrototype for <' + kpmGetDocumentUrl(ADocument) + '>');
            var ScriptElement = ADocument.createElement("SCRIPT");
            ScriptElement.appendChild(ADocument.createTextNode("try"));
            ScriptElement.appendChild(ADocument.createTextNode("{"));
            ScriptElement.appendChild(ADocument.createTextNode("   document.objHTMLFormElementPrototype = new Object();"));
            ScriptElement.appendChild(ADocument.createTextNode("   document.objHTMLFormElementPrototype.submit = HTMLFormElement.prototype.submit;"));
            ScriptElement.appendChild(ADocument.createTextNode("   HTMLFormElement.prototype.submit = function(AEvent)"));
            ScriptElement.appendChild(ADocument.createTextNode("   {"));
            // >> ScriptElement.appendChild(ADocument.createTextNode("      var AForm = AEvent ? AEvent.target : this;"));
            // >> ScriptElement.appendChild(ADocument.createTextNode("      alert('Submitting form through JavaScript to: <' + AForm.action + '>');"));
            ScriptElement.appendChild(ADocument.createTextNode("      var kpmEvent = document.createEvent('Event');"));
            ScriptElement.appendChild(ADocument.createTextNode("      kpmEvent.initEvent('kpm_submit', true, true);"));
            ScriptElement.appendChild(ADocument.createTextNode("      this.dispatchEvent(kpmEvent);"));
            ScriptElement.appendChild(ADocument.createTextNode("      document.objHTMLFormElementPrototype.submit.apply(this);"));
            ScriptElement.appendChild(ADocument.createTextNode("   };"));
            ScriptElement.appendChild(ADocument.createTextNode("}"));
            ScriptElement.appendChild(ADocument.createTextNode("catch(e)"));
            ScriptElement.appendChild(ADocument.createTextNode("{"));
            ScriptElement.appendChild(ADocument.createTextNode("  alert('Error during attach to HTMLFormElement.prototype.submit: ' + ErrorMessage);"));
            ScriptElement.appendChild(ADocument.createTextNode("}"));	 
			// force attaching to 'submit' event on ASP .NET sites 
            ScriptElement.appendChild(ADocument.createTextNode("kpm_addToPostBack = function(func)"));
            ScriptElement.appendChild(ADocument.createTextNode("{"));
            ScriptElement.appendChild(ADocument.createTextNode("  if (typeof __doPostBack == 'function')"));
            ScriptElement.appendChild(ADocument.createTextNode("  {"));
            ScriptElement.appendChild(ADocument.createTextNode("    var old__doPostBack = __doPostBack;"));
            ScriptElement.appendChild(ADocument.createTextNode("    __doPostBack = function(t, a)"));
            ScriptElement.appendChild(ADocument.createTextNode("      {"));
            ScriptElement.appendChild(ADocument.createTextNode("        if (func(t, a)) old__doPostBack(t, a);"));
            ScriptElement.appendChild(ADocument.createTextNode("      }"));
            ScriptElement.appendChild(ADocument.createTextNode("  }"));
            ScriptElement.appendChild(ADocument.createTextNode("};"));
            ScriptElement.appendChild(ADocument.createTextNode("kpm_addToPostBack(function(t,a) {"));            
            ScriptElement.appendChild(ADocument.createTextNode("  var kpmEvent = document.createEvent('Event');"));
            ScriptElement.appendChild(ADocument.createTextNode("  kpmEvent.initEvent('kpm_submit', true, true);"));
            ScriptElement.appendChild(ADocument.createTextNode("  this.dispatchEvent(kpmEvent);"));                        
            ScriptElement.appendChild(ADocument.createTextNode("  return true;"));
            ScriptElement.appendChild(ADocument.createTextNode("});"));
            ScriptElement.setAttribute("id", "objHTMLFormElementPrototype");
            ParentElement.appendChild(ScriptElement);
          }
        }
      }
      /*
      // not-working code:
      eval(
        "try" +
        "{" +
        "  if (HTMLFormElement.prototype._submit == undefined ||" +
        "      !HTMLFormElement.prototype._submit)" +
        "  {" +
        "    HTMLFormElement.prototype._submit = HTMLFormElement.prototype.submit;" + 
        "    HTMLFormElement.prototype.submit = function(AEvent)" +
        "    {" +
        // >>  "      var AForm = AEvent ? AEvent.target : this;" +
        // >>  "      alert('Submitting form through JavaScript to: <' + AForm.action + '>');" +
        "      var kpmEvent = document.createEvent('Event');" + 
        "      kpmEvent.initEvent('kpm_submit', true, true);" +
        "      this.dispatchEvent(kpmEvent);" +
        "      this._submit();" +
        "    }" +
        "  }" +
        "}" + 
        "catch (ErrorMessage)" +
        "{" +
        "  alert('Error during attach to HTMLFormElement.prototype.submit: ' + ErrorMessage);" +
        "}", ADocument
      );
      */
    }
  }
  catch (ErrorMessage)
  {
    kpmLog.logError('kpmInstallHTMLFormElementPrototypeForWindow() Error: ' + ErrorMessage);
  }  
}

// << Content Core Script ======================================================